from xpxchain import client
from xpxchain import models
from tests import harness
from tests import config
import binascii
import hashlib
import os
from binascii import hexlify
import typing
import asyncio
from tests import helper

send_funds = helper.send_funds
announce = helper.announce
announce_partial = helper.announce_partial
announce_cosignature = helper.announce_cosignature

_1 = config.divisibility * 1
_10 = _1 * 10
_100 = _1 * 100
_1000 = _1 * 1000
_10000 = _1 * 10000


class TestT1Http(harness.TestCase):

    t1: typing.Sequence[models.Account]
    t2: typing.Sequence[models.Account]
    t3: typing.Sequence[models.Account]
    t4: typing.Sequence[models.Account]
    t5: typing.Sequence[models.Account]
    t6: typing.Sequence[models.Account]
    hashes: typing.Sequence[str] = []

    @classmethod
    def setUpClass(cls):

        print("Setting up tests... ")

        cls.t1 = [models.Account.generate_new_account(config.network_type, entropy=lambda x: os.urandom(32)) for i in range(2)]
        cls.t2 = [models.Account.generate_new_account(config.network_type, entropy=lambda x: os.urandom(32)) for i in range(19)]
        cls.t3 = [models.Account.generate_new_account(config.network_type, entropy=lambda x: os.urandom(32)) for i in range(3)]
        cls.t4 = [models.Account.generate_new_account(config.network_type, entropy=lambda x: os.urandom(32)) for i in range(2)]
        cls.t5 = [models.Account.generate_new_account(config.network_type, entropy=lambda x: os.urandom(32)) for i in range(3)]
        cls.t6 = [models.Account.generate_new_account(config.network_type, entropy=lambda x: os.urandom(32)) for i in range(1)]

        loop = asyncio.get_event_loop()
        cls.hashes = loop.run_until_complete(helper.prepare(
            [send_funds(config.tester, account, _1) for account in cls.t1]
            + [send_funds(config.tester, account, _10) for account in cls.t2]
            + [send_funds(config.tester, account, _100) for account in cls.t3]
            + [send_funds(config.tester, account, _10000 + _100) for account in cls.t4]
            + [send_funds(config.tester, account, 2 * _10000 + _100) for account in cls.t5]
            + [send_funds(config.tester, account, 3 * _10000 + _100) for account in cls.t6],
            max_run=10
        ))

    # TESTS

    def test_get_transaction(self):
        with client.TransactionHTTP(config.ENDPOINT, network_type=config.network_type) as http:
            reply = http.get_transaction(self.hashes[0])
            self.assertEqual(isinstance(reply, models.TransferTransaction), True)
            self.assertEqual(reply.transaction_info.hash, self.hashes[0])

    def test_get_transactions(self):
        with client.TransactionHTTP(config.ENDPOINT, network_type=config.network_type) as http:
            reply = http.get_transactions(self.hashes[:2])
            self.assertEqual(len(reply), len(self.hashes[:2]))
            self.assertEqual(isinstance(reply[0], models.TransferTransaction), True)
            for tx in reply:
                self.assertEqual(tx.transaction_info.hash in self.hashes, True)

    def test_get_transaction_status(self):
        with client.TransactionHTTP(config.ENDPOINT, network_type=config.network_type) as http:
            reply = http.get_transaction_status(self.hashes[0])
            self.assertEqual(isinstance(reply, models.TransactionStatus), True)
            self.assertEqual(reply.hash, self.hashes[0])

    def test_get_transaction_statuses(self):
        with client.TransactionHTTP(config.ENDPOINT, network_type=config.network_type) as http:
            reply = http.get_transaction_statuses(self.hashes[:2])
            self.assertEqual(len(reply), len(self.hashes[:2]))
            self.assertEqual(isinstance(reply[0], models.TransactionStatus), True)
            for hash in reply:
                self.assertEqual(hash.hash in self.hashes, True)

#    def test_transfer_transaction(self):
#        self.send_funds(config.tester, alice, _10)
#
    async def test_message_transaction(self):
        alice = self.t2.pop()
        bob = models.Account.generate_new_account(config.network_type, entropy=lambda x: os.urandom(32))

        message = models.PlainMessage(b'Hello world')

        tx = models.TransferTransaction.create(
            deadline=models.Deadline.create(),
            recipient=bob.address,
            network_type=config.network_type,
            message=message
        )

        signed_tx = tx.sign_with(alice, config.gen_hash)

        tx = await announce(signed_tx)
        self.assertEqual(isinstance(tx, models.TransferTransaction), True)
        self.assertEqual(tx.recipient, bob.address)
        self.assertEqual(tx.message, message)

    async def test_account_link_transaction(self):
        alice = self.t2.pop()
        bob = models.Account.generate_new_account(config.network_type, entropy=lambda x: os.urandom(32))

        tx = models.AccountLinkTransaction.create(
            deadline=models.Deadline.create(),
            remote_account_key=bob.public_key,
            link_action=models.LinkAction.LINK,
            network_type=config.network_type,
        )

        signed_tx = tx.sign_with(alice, config.gen_hash)

        tx = await announce(signed_tx)
        self.assertEqual(isinstance(tx, models.AccountLinkTransaction), True)
        self.assertEqual(tx.remote_account_key, bob.public_key.upper())
        self.assertEqual(tx.link_action, models.LinkAction.LINK)

        tx = models.AccountLinkTransaction.create(
            deadline=models.Deadline.create(),
            remote_account_key=bob.public_key,
            link_action=models.LinkAction.UNLINK,
            network_type=config.network_type,
        )

        signed_tx = tx.sign_with(alice, config.gen_hash)

        tx = await announce(signed_tx)
        self.assertEqual(isinstance(tx, models.AccountLinkTransaction), True)
        self.assertEqual(tx.remote_account_key, bob.public_key.upper())
        self.assertEqual(tx.link_action, models.LinkAction.UNLINK)

    async def test_modify_account_property_address_transaction(self):
        alice = self.t2.pop()
        bob = models.Account.generate_new_account(config.network_type, entropy=lambda x: os.urandom(32))

        for property_type in [models.PropertyType.ALLOW_ADDRESS, models.PropertyType.BLOCK_ADDRESS]:
            for modification_type in [models.PropertyModificationType.ADD, models.PropertyModificationType.REMOVE]:

                tx = models.ModifyAccountPropertyAddressTransaction.create(
                    deadline=models.Deadline.create(),
                    network_type=config.network_type,
                    property_type=property_type,
                    modifications=[models.AccountPropertyModification(modification_type, bob.address)]
                )

                signed_tx = tx.sign_with(alice, config.gen_hash)

                tx = await announce(signed_tx)
                self.assertEqual(isinstance(tx, models.ModifyAccountPropertyAddressTransaction), True)
                self.assertEqual(len(tx.modifications), 1)
                self.assertEqual(tx.property_type, property_type)
                self.assertEqual(tx.modifications[0].modification_type, modification_type)
                self.assertEqual(tx.modifications[0].value, bob.address)

    async def test_modify_account_property_mosaic_transaction(self):
        alice = self.t2.pop()

        for property_type in [models.PropertyType.ALLOW_MOSAIC, models.PropertyType.BLOCK_MOSAIC]:
            for modification_type in [models.PropertyModificationType.ADD, models.PropertyModificationType.REMOVE]:

                tx = models.ModifyAccountPropertyMosaicTransaction.create(
                    deadline=models.Deadline.create(),
                    network_type=config.network_type,
                    property_type=property_type,
                    modifications=[models.AccountPropertyModification(modification_type, config.mosaic_id)]
                )

                signed_tx = tx.sign_with(alice, config.gen_hash)

                tx = await announce(signed_tx)
                self.assertEqual(isinstance(tx, models.ModifyAccountPropertyMosaicTransaction), True)
                self.assertEqual(len(tx.modifications), 1)
                self.assertEqual(tx.property_type, property_type)
                self.assertEqual(tx.modifications[0].modification_type, modification_type)
                self.assertEqual(tx.modifications[0].value, config.mosaic_id)

    async def test_modify_account_property_entity_type_transaction(self):
        alice = self.t2.pop()

        tx_type = models.TransactionType.AGGREGATE_COMPLETE

        for property_type in [models.PropertyType.BLOCK_TRANSACTION]:
            for modification_type in [models.PropertyModificationType.ADD, models.PropertyModificationType.REMOVE]:

                tx = models.ModifyAccountPropertyEntityTypeTransaction.create(
                    deadline=models.Deadline.create(),
                    network_type=config.network_type,
                    property_type=property_type,
                    modifications=[models.AccountPropertyModification(modification_type, tx_type)]
                )

                signed_tx = tx.sign_with(alice, config.gen_hash)

                tx = await announce(signed_tx)
                self.assertEqual(isinstance(tx, models.ModifyAccountPropertyEntityTypeTransaction), True)
                self.assertEqual(len(tx.modifications), 1)
                self.assertEqual(tx.property_type, property_type)
                self.assertEqual(tx.modifications[0].modification_type, modification_type)
                self.assertEqual(tx.modifications[0].value, tx_type)

    async def test_register_namespace_transaction(self):
        alice = self.t6.pop()

        namespace_l1 = 'foo' + hexlify(os.urandom(4)).decode('utf-8')
        namespace_l2 = 'bar'
        namespace_name = namespace_l1 + '.' + namespace_l2

        n1, n2 = await helper.create_namespace(alice, namespace_name)
        self.assertEqual(n1.name, namespace_l1)
        self.assertEqual(n2.name, namespace_l2)
        self.assertEqual(isinstance(n2.parent_id, models.NamespaceId), True)

        # Link address alias
        for action_type in [models.AliasActionType.LINK, models.AliasActionType.UNLINK]:
            tx = models.AddressAliasTransaction.create(
                deadline=models.Deadline.create(),
                network_type=config.network_type,
                max_fee=1,
                action_type=action_type,
                namespace_id=n2.namespace_id,
                address=alice.address
            )

            signed_tx = tx.sign_with(alice, config.gen_hash)

            tx = await announce(signed_tx)
            self.assertEqual(isinstance(tx, models.AddressAliasTransaction), True)
            self.assertEqual(tx.action_type, action_type)
            self.assertEqual(tx.address, alice.address)

        mosaic_id = await helper.create_mosaic(alice, 1)
        self.assertEqual(isinstance(mosaic_id, models.MosaicId), True)

        # Link mosaic alias
        for action_type in [models.AliasActionType.LINK, models.AliasActionType.UNLINK]:
            tx = models.MosaicAliasTransaction.create(
                deadline=models.Deadline.create(),
                network_type=config.network_type,
                max_fee=1,
                action_type=action_type,
                namespace_id=n2.namespace_id,
                mosaic_id=mosaic_id,
            )

            signed_tx = tx.sign_with(alice, config.gen_hash)

            tx = await announce(signed_tx)
            self.assertEqual(isinstance(tx, models.MosaicAliasTransaction), True)
            self.assertEqual(tx.mosaic_id, mosaic_id)
            self.assertEqual(tx.action_type, action_type)

    async def test_secret_lock_transaction(self):
        alice = self.t2.pop()
        bob = models.Account.generate_new_account(config.network_type, entropy=lambda x: os.urandom(32))

        random_bytes = os.urandom(20)
        h = hashlib.sha3_256(random_bytes)
        secret = binascii.hexlify(h.digest()).decode('utf-8').upper()
        proof = binascii.hexlify(random_bytes).decode('utf-8').upper()

        tx = models.SecretLockTransaction.create(
            deadline=models.Deadline.create(),
            network_type=config.network_type,
            mosaic=models.Mosaic(config.mosaic_id, _1),
            duration=60,
            hash_type=models.HashType.SHA3_256,
            secret=secret,
            recipient=bob.address,
        )

        signed_tx = tx.sign_with(alice, config.gen_hash)

        tx = await announce(signed_tx)
        self.assertEqual(isinstance(tx, models.SecretLockTransaction), True)
        self.assertEqual(tx.recipient, bob.address)
        self.assertEqual(tx.secret, secret)

        tx = models.SecretProofTransaction.create(
            deadline=models.Deadline.create(),
            network_type=config.network_type,
            hash_type=models.HashType.SHA3_256,
            secret=secret,
            proof=proof,
            recipient=bob.address,
        )

        signed_tx = tx.sign_with(alice, config.gen_hash)

        tx = await announce(signed_tx)
        self.assertEqual(isinstance(tx, models.SecretProofTransaction), True)
        self.assertEqual(tx.recipient, bob.address)
        self.assertEqual(tx.secret, secret)

    async def test_aggregate_transaction_with_cosigners(self):
        alice = self.t3.pop()
        bob = self.t2.pop()

        alice_to_bob = models.TransferTransaction.create(
            deadline=models.Deadline.create(),
            recipient=bob.address,
            mosaics=[models.Mosaic(config.mosaic_id, _1)],
            network_type=config.network_type,
        )

        bob_to_alice = models.TransferTransaction.create(
            deadline=models.Deadline.create(),
            recipient=alice.address,
            mosaics=[models.Mosaic(config.mosaic_id, _1)],
            network_type=config.network_type,
        )

        tx = models.AggregateTransaction.create_complete(
            deadline=models.Deadline.create(),
            inner_transactions=[bob_to_alice.to_aggregate(bob), alice_to_bob.to_aggregate(alice)],
            network_type=config.network_type,
        )

        signed_tx = tx.sign_transaction_with_cosignatories(alice, config.gen_hash, [bob])

        tx = await announce(signed_tx)
        self.assertEqual(isinstance(tx, models.AggregateTransaction), True)
        self.assertEqual(len(tx.inner_transactions), 2)
        self.assertEqual(tx.inner_transactions[0].recipient, alice.address)
        self.assertEqual(tx.inner_transactions[1].recipient, bob.address)

    async def test_aggregate_bonded_transaction(self):
        alice = self.t3.pop()
        bob = self.t2.pop()

        alice_to_bob = models.TransferTransaction.create(
            deadline=models.Deadline.create(),
            recipient=bob.address,
            mosaics=[models.Mosaic(config.mosaic_id, _1)],
            network_type=config.network_type,
        )

        bob_to_alice = models.TransferTransaction.create(
            deadline=models.Deadline.create(),
            recipient=alice.address,
            mosaics=[models.Mosaic(config.mosaic_id, _1)],
            network_type=config.network_type,
        )

        bonded = models.AggregateTransaction.create_bonded(
            deadline=models.Deadline.create(),
            inner_transactions=[bob_to_alice.to_aggregate(bob), alice_to_bob.to_aggregate(alice)],
            network_type=config.network_type,
        )

        signed_bonded = bonded.sign_transaction_with_cosignatories(alice, config.gen_hash)

        lock = models.LockFundsTransaction.create(
            deadline=models.Deadline.create(),
            network_type=config.network_type,
            mosaic=models.Mosaic(config.mosaic_id, _10),
            duration=60,
            signed_transaction=signed_bonded,
        )

        signed_lock = lock.sign_with(alice, config.gen_hash)

        tx = await announce(signed_lock)
        self.assertEqual(isinstance(tx, models.LockFundsTransaction), True)

        tx = await announce_partial(signed_bonded)
        self.assertEqual(isinstance(tx, models.AggregateTransaction), True)

        with client.AccountHTTP(config.ENDPOINT, network_type=config.network_type) as http:
            reply = http.aggregate_bonded_transactions(bob)
            self.assertEqual(isinstance(reply[0], models.Transaction), True)
            self.assertEqual(isinstance(reply[0], models.AggregateTransaction), True)
            self.assertEqual(isinstance(reply[0], models.AggregateBondedTransaction), True)

            signed_cosig = models.CosignatureTransaction.create(reply[0]).sign_with(bob)

        tx = await announce_cosignature(signed_cosig)
        self.assertEqual(isinstance(tx, models.AggregateTransaction), True)

    async def test_create_multisig_and_send_funds(self):
        multisig = self.t3.pop()
        alice = self.t2.pop()
        bob = models.Account.generate_new_account(config.network_type, entropy=lambda x: os.urandom(32))

        change_to_multisig = models.ModifyMultisigAccountTransaction.create(
            deadline=models.Deadline.create(),
            min_approval_delta=2,
            min_removal_delta=1,
            modifications=[
                models.MultisigCosignatoryModification.create(alice, models.MultisigCosignatoryModificationType.ADD),
                models.MultisigCosignatoryModification.create(bob, models.MultisigCosignatoryModificationType.ADD),
            ],
            network_type=config.network_type,
        )

        tx = models.AggregateTransaction.create_complete(
            deadline=models.Deadline.create(),
            inner_transactions=[change_to_multisig.to_aggregate(multisig)],
            network_type=config.network_type,
        )

        signed_tx = tx.sign_transaction_with_cosignatories(multisig, config.gen_hash, [alice, bob])

        tx = await announce(signed_tx)
        self.assertEqual(isinstance(tx, models.AggregateTransaction), True)

        multisig_to_tester = models.TransferTransaction.create(
            deadline=models.Deadline.create(),
            recipient=config.tester.address,
            mosaics=[models.Mosaic(config.mosaic_id, _1)],
            network_type=config.network_type,
        )

        tx = models.AggregateTransaction.create_complete(
            deadline=models.Deadline.create(),
            inner_transactions=[multisig_to_tester.to_aggregate(multisig)],
            network_type=config.network_type,
        )

        signed_tx = tx.sign_transaction_with_cosignatories(alice, config.gen_hash, [bob])

        tx = await announce(signed_tx)
        self.assertEqual(isinstance(tx, models.AggregateTransaction), True)

    async def test_multisig_account_info(self):
        alice = models.Account.generate_new_account(config.network_type, entropy=lambda x: os.urandom(32))
        bob = models.Account.generate_new_account(config.network_type, entropy=lambda x: os.urandom(32))
        multisig = self.t2.pop()

        change_to_multisig = models.ModifyMultisigAccountTransaction.create(
            deadline=models.Deadline.create(),
            min_approval_delta=2,
            min_removal_delta=1,
            modifications=[
                models.MultisigCosignatoryModification.create(alice, models.MultisigCosignatoryModificationType.ADD),
                models.MultisigCosignatoryModification.create(bob, models.MultisigCosignatoryModificationType.ADD),
            ],
            network_type=config.network_type,
        )

        tx = models.AggregateTransaction.create_complete(
            deadline=models.Deadline.create(),
            inner_transactions=[change_to_multisig.to_aggregate(multisig)],
            network_type=config.network_type,
        )

        signed_tx = tx.sign_transaction_with_cosignatories(multisig, config.gen_hash, [alice, bob])

        tx = await announce(signed_tx)
        self.assertEqual(isinstance(tx, models.AggregateTransaction), True)

        with client.AccountHTTP(config.ENDPOINT, network_type=config.network_type) as http:
            info = http.get_multisig_account_info(multisig.address)
            self.assertEqual(isinstance(info, models.MultisigAccountInfo), True)

    async def test_multisig_account_graph_info(self):
        alice = models.Account.generate_new_account(config.network_type, entropy=lambda x: os.urandom(32))
        bob = models.Account.generate_new_account(config.network_type, entropy=lambda x: os.urandom(32))
        mike = models.Account.generate_new_account(config.network_type, entropy=lambda x: os.urandom(32))
        multisig = self.t2.pop()
        multisig2 = self.t2.pop()

        change_to_multisig = models.ModifyMultisigAccountTransaction.create(
            deadline=models.Deadline.create(),
            min_approval_delta=2,
            min_removal_delta=1,
            modifications=[
                models.MultisigCosignatoryModification.create(alice, models.MultisigCosignatoryModificationType.ADD),
                models.MultisigCosignatoryModification.create(bob, models.MultisigCosignatoryModificationType.ADD),
            ],
            network_type=config.network_type,
        )

        tx = models.AggregateTransaction.create_complete(
            deadline=models.Deadline.create(),
            inner_transactions=[change_to_multisig.to_aggregate(multisig)],
            network_type=config.network_type,
        )

        signed_tx = tx.sign_transaction_with_cosignatories(multisig, config.gen_hash, [alice, bob])

        tx = await announce(signed_tx)
        self.assertEqual(isinstance(tx, models.AggregateTransaction), True)

        change_to_multisig = models.ModifyMultisigAccountTransaction.create(
            deadline=models.Deadline.create(),
            min_approval_delta=2,
            min_removal_delta=1,
            modifications=[
                models.MultisigCosignatoryModification.create(mike, models.MultisigCosignatoryModificationType.ADD),
                models.MultisigCosignatoryModification.create(multisig, models.MultisigCosignatoryModificationType.ADD),
            ],
            network_type=config.network_type,
        )

        tx = models.AggregateTransaction.create_complete(
            deadline=models.Deadline.create(),
            inner_transactions=[change_to_multisig.to_aggregate(multisig2)],
            network_type=config.network_type,
        )

        signed_tx = tx.sign_transaction_with_cosignatories(multisig2, config.gen_hash, [alice, bob, mike])

        tx = await announce(signed_tx)
        self.assertEqual(isinstance(tx, models.AggregateTransaction), True)

        with client.AccountHTTP(config.ENDPOINT, network_type=config.network_type) as http:
            info = http.get_multisig_account_graph_info(multisig.address)
            self.assertEqual(isinstance(info, models.MultisigAccountGraphInfo), True)

    async def test_account_names(self):
        alice = self.t5.pop()

        namespace_name = 'foo' + hexlify(os.urandom(4)).decode('utf-8') + '.bar'
        n1, n2 = await helper.create_namespace(alice, namespace_name)

        # Link address alias
        tx = models.AddressAliasTransaction.create(
            deadline=models.Deadline.create(),
            network_type=config.network_type,
            max_fee=1,
            action_type=models.AliasActionType.LINK,
            namespace_id=n2.namespace_id,
            address=alice.address
        )

        signed_tx = tx.sign_with(alice, config.gen_hash)

        tx = await announce(signed_tx)
        self.assertEqual(isinstance(tx, models.AddressAliasTransaction), True)
        self.assertEqual(tx.action_type, models.AliasActionType.LINK)
        self.assertEqual(tx.address, alice.address)

        with client.AccountHTTP(config.ENDPOINT, network_type=config.network_type) as http:
            info = http.get_account_names([alice.address])
            self.assertEqual(len(info), 1)
            self.assertEqual(isinstance(info[0], models.AccountNames), True)
            self.assertEqual(len(info[0].names), 1)
            self.assertEqual(info[0].names[0], namespace_name)

    async def test_account_transactions(self):
        alice = self.t1.pop()

        with client.AccountHTTP(config.ENDPOINT, network_type=config.network_type) as http:
            info = http.transactions(alice)
            self.assertEqual(len(info), 1)
            self.assertEqual(len(info[0].mosaics), 1)
            self.assertEqual(info[0].mosaics[0], models.Mosaic(config.mosaic_id, _1))
            self.assertEqual(info[0].signer.public_key, config.tester.public_key.upper())

    async def test_incoming_transactions(self):
        alice = self.t1.pop()

        with client.AccountHTTP(config.ENDPOINT, network_type=config.network_type) as http:
            info = http.incoming_transactions(alice)
            self.assertEqual(len(info), 1)
            self.assertEqual(len(info[0].mosaics), 1)
            self.assertEqual(info[0].mosaics[0], models.Mosaic(config.mosaic_id, _1))
            self.assertEqual(info[0].signer.public_key, config.tester.public_key.upper())

    async def test_outgoing_transactions(self):
        alice = self.t2.pop()
        bob = models.Account.generate_new_account(config.network_type, entropy=lambda x: os.urandom(32))

        await send_funds(alice, bob, _1, quiet=True)

        with client.AccountHTTP(config.ENDPOINT, network_type=config.network_type) as http:
            info = http.outgoing_transactions(alice)
            self.assertEqual(len(info), 1)
            self.assertEqual(len(info[0].mosaics), 1)
            self.assertEqual(info[0].mosaics[0], models.Mosaic(config.mosaic_id, _1))
            self.assertEqual(info[0].recipient, bob.address)

    async def test_account_properties(self):
        alice = self.t2.pop()
        bob = models.Account.generate_new_account(config.network_type, entropy=lambda x: os.urandom(32))
        pete = self.t2.pop()

        # ALLOW_ADDRESS
        tx = models.ModifyAccountPropertyAddressTransaction.create(
            deadline=models.Deadline.create(),
            network_type=config.network_type,
            property_type=models.PropertyType.ALLOW_ADDRESS,
            modifications=[models.AccountPropertyModification(models.PropertyModificationType.ADD, bob.address)]
        )

        signed_tx = tx.sign_with(alice, config.gen_hash)

        tx = await announce(signed_tx)

        # ALLOW_MOSAIC
        tx = models.ModifyAccountPropertyMosaicTransaction.create(
            deadline=models.Deadline.create(),
            network_type=config.network_type,
            property_type=models.PropertyType.ALLOW_MOSAIC,
            modifications=[models.AccountPropertyModification(models.PropertyModificationType.ADD, config.mosaic_id)]
        )

        signed_tx = tx.sign_with(alice, config.gen_hash)

        tx = await announce(signed_tx)

        # BLOCK_ADDRESS
        tx = models.ModifyAccountPropertyAddressTransaction.create(
            deadline=models.Deadline.create(),
            network_type=config.network_type,
            property_type=models.PropertyType.BLOCK_ADDRESS,
            modifications=[models.AccountPropertyModification(models.PropertyModificationType.ADD, bob.address)]
        )

        signed_tx = tx.sign_with(pete, config.gen_hash)

        tx = await announce(signed_tx)

        # BLOCK_MOSAIC
        tx = models.ModifyAccountPropertyMosaicTransaction.create(
            deadline=models.Deadline.create(),
            network_type=config.network_type,
            property_type=models.PropertyType.BLOCK_MOSAIC,
            modifications=[models.AccountPropertyModification(models.PropertyModificationType.ADD, config.mosaic_id)]
        )

        signed_tx = tx.sign_with(pete, config.gen_hash)

        tx = await announce(signed_tx)

        # BLOCK_TRANSACTION
        tx = models.ModifyAccountPropertyEntityTypeTransaction.create(
            deadline=models.Deadline.create(),
            network_type=config.network_type,
            property_type=models.PropertyType.BLOCK_TRANSACTION,
            modifications=[models.AccountPropertyModification(models.PropertyModificationType.ADD, models.TransactionType.AGGREGATE_COMPLETE)]
        )

        signed_tx = tx.sign_with(pete, config.gen_hash)

        tx = await announce(signed_tx)

        with client.AccountHTTP(config.ENDPOINT, network_type=config.network_type) as http:
            info = http.get_account_properties(alice.address)
            self.assertEqual(len(info.properties), 3)
            self.assertEqual(info.properties[0].property_type, models.PropertyType.ALLOW_ADDRESS)
            self.assertEqual(len(info.properties[0].values), 1)
            self.assertEqual(info.properties[0].values[0], bob.address)
            self.assertEqual(info.properties[1].property_type, models.PropertyType.ALLOW_MOSAIC)
            self.assertEqual(len(info.properties[1].values), 1)
            self.assertEqual(info.properties[1].values[0], config.mosaic_id)
            self.assertEqual(info.properties[2].property_type, models.PropertyType.BLOCK_TRANSACTION)
            self.assertEqual(len(info.properties[2].values), 0)

            info = http.get_account_properties(pete.address)
            self.assertEqual(len(info.properties), 3)
            self.assertEqual(info.properties[0].property_type, models.PropertyType.BLOCK_ADDRESS)
            self.assertEqual(len(info.properties[0].values), 1)
            self.assertEqual(info.properties[0].values[0], bob.address)
            self.assertEqual(info.properties[1].property_type, models.PropertyType.BLOCK_MOSAIC)
            self.assertEqual(len(info.properties[1].values), 1)
            self.assertEqual(info.properties[1].values[0], config.mosaic_id)
            self.assertEqual(info.properties[2].property_type, models.PropertyType.BLOCK_TRANSACTION)
            self.assertEqual(len(info.properties[2].values), 1)
            self.assertEqual(info.properties[2].values[0], models.TransactionType.AGGREGATE_COMPLETE)

            info = http.get_accounts_properties([alice.address, pete.address])
            self.assertEqual(len(info), 2)

            for i in info:
                if (i.address == alice.address):
                    self.assertEqual(len(i.properties), 3)
                    self.assertEqual(i.properties[0].property_type, models.PropertyType.ALLOW_ADDRESS)
                    self.assertEqual(len(i.properties[0].values), 1)
                    self.assertEqual(i.properties[0].values[0], bob.address)
                    self.assertEqual(i.properties[1].property_type, models.PropertyType.ALLOW_MOSAIC)
                    self.assertEqual(len(i.properties[1].values), 1)
                    self.assertEqual(i.properties[1].values[0], config.mosaic_id)
                    self.assertEqual(i.properties[2].property_type, models.PropertyType.BLOCK_TRANSACTION)
                    self.assertEqual(len(i.properties[2].values), 0)
                    self.assertEqual(len(i.properties), 3)
                elif (i.address == pete.address):
                    self.assertEqual(i.properties[0].property_type, models.PropertyType.BLOCK_ADDRESS)
                    self.assertEqual(len(i.properties[0].values), 1)
                    self.assertEqual(i.properties[0].values[0], bob.address)
                    self.assertEqual(i.properties[1].property_type, models.PropertyType.BLOCK_MOSAIC)
                    self.assertEqual(len(i.properties[1].values), 1)
                    self.assertEqual(i.properties[1].values[0], config.mosaic_id)
                    self.assertEqual(i.properties[2].property_type, models.PropertyType.BLOCK_TRANSACTION)
                    self.assertEqual(len(i.properties[2].values), 1)
                    self.assertEqual(i.properties[2].values[0], models.TransactionType.AGGREGATE_COMPLETE)
                else:
                    raise Exception(f'Unknown address{i.address.address}')

    async def test_aggregate_bonded_transactions(self):
        alice = self.t2.pop()

        # TODO: Add bonded transaction
        with client.AccountHTTP(config.ENDPOINT, network_type=config.network_type) as http:
            info = http.aggregate_bonded_transactions(alice)
            self.assertEqual(len(info), 0)

    async def test_unconfirmed_transactions(self):
        alice = self.t2.pop()

        # TODO: Add unconfirmed transaction
        with client.AccountHTTP(config.ENDPOINT, network_type=config.network_type) as http:
            info = http.unconfirmed_transactions(alice)
            self.assertEqual(len(info), 0)

    async def test_get_account_info(self):
        with client.AccountHTTP(config.ENDPOINT, network_type=config.network_type) as http:
            info = http.get_account_info(config.tester.address)
            self.assertEqual(info.public_key, config.tester.public_key.upper()),

    async def test_get_accountis_info(self):
        with client.AccountHTTP(config.ENDPOINT, network_type=config.network_type) as http:
            info = http.get_accounts_info([config.tester.address])
            self.assertEqual(len(info), 1)
            self.assertEqual(info[0].public_key, config.tester.public_key.upper())

    def test_get_merkle_by_hash_in_block(self):
        with client.BlockchainHTTP(config.ENDPOINT, network_type=config.network_type) as http:
            reply = http.get_block_transactions(1)
            tx_hash = reply[0].transaction_info.hash

            info = http.get_merkle_by_hash_in_block(1, tx_hash)
            self.assertEqual(isinstance(info, models.MerkleProofInfo), True)
            self.assertEqual(len(info.merkle_path) > 0, True)
            self.assertEqual(isinstance(info.merkle_path[0], models.MerklePathItem), True)

    def test_get_blocks_by_height_with_limit(self):
        with client.BlockchainHTTP(config.ENDPOINT, network_type=config.network_type) as http:
            info = http.get_blocks_by_height_with_limit(25, 25)
            self.assertEqual(len(info), 25)
            self.assertEqual(isinstance(info[0], models.BlockInfo), True)

    def test_get_block_by_height(self):
        with client.BlockchainHTTP(config.ENDPOINT, network_type=config.network_type) as http:
            info = http.get_block_by_height(25)
            self.assertEqual(isinstance(info, models.BlockInfo), True)

    def test_get_block_transactions(self):
        with client.BlockchainHTTP(config.ENDPOINT, network_type=config.network_type) as http:
            info = http.get_block_transactions(1)
            self.assertEqual(len(info) > 0, True)
            self.assertEqual(isinstance(info[0], models.Transaction), True)
            self.assertEqual(info[0].signer.public_key, config.nemesis.public_key.upper())

    def test_get_block_receipts(self):
        with client.BlockchainHTTP(config.ENDPOINT, network_type=config.network_type) as http:
            info = http.get_block_receipts(25)
            self.assertEqual(isinstance(info, models.Statements), True)
            self.assertEqual(len(info.transaction_statements), 1)
            self.assertEqual(isinstance(info.transaction_statements[0], models.TransactionStatement), True)
            self.assertEqual(len(info.transaction_statements[0].receipts) > 0, True)
            self.assertEqual(isinstance(info.transaction_statements[0].receipts[0], models.BalanceChangeReceipt), True)

    def test_get_diagnostic_storage(self):
        with client.BlockchainHTTP(config.ENDPOINT, network_type=config.network_type) as http:
            info = http.get_diagnostic_storage()
            self.assertEqual(isinstance(info, models.BlockchainStorageInfo), True)

    def test_get_diagnostic_server(self):
        with client.BlockchainHTTP(config.ENDPOINT, network_type=config.network_type) as http:
            info = http.get_diagnostic_server()
            self.assertEqual(isinstance(info, models.BlockchainServerInfo), True)

    def test_get_blockchain_height(self):
        with client.BlockchainHTTP(config.ENDPOINT, network_type=config.network_type) as http:
            info = http.get_blockchain_height()
            self.assertEqual(isinstance(info, int), True)
            self.assertEqual(info >= 1, True)

    def test_get_blockchain_score(self):
        with client.BlockchainHTTP(config.ENDPOINT, network_type=config.network_type) as http:
            info = http.get_blockchain_score()
            self.assertEqual(isinstance(info, models.BlockchainScore), True)

    def test_get_config(self):
        with client.ConfigHTTP(config.ENDPOINT, network_type=config.network_type) as http:
            info = http.get_config(1)
            self.assertEqual(isinstance(info, models.CatapultConfig), True)

    def test_get_upgrade(self):
        with client.ConfigHTTP(config.ENDPOINT, network_type=config.network_type) as http:
            info = http.get_upgrade(1)
            self.assertEqual(isinstance(info, models.CatapultUpgrade), True)

    async def test_new_block(self):
        async with client.Listener(f'{config.ENDPOINT}/ws', network_type=config.network_type) as listener:
            await listener.new_block()

            async for m in listener:
                self.assertEqual(m.channel_name, 'block')
                self.assertEqual(isinstance(m.message, models.BlockInfo), True)
                self.assertEqual(m.message.height >= 1, True)
                break

    async def test_modify_address_metadata_transaction(self):
        alice = self.t2.pop()

        metadata_key = 'foo' + hexlify(os.urandom(4)).decode('ascii')

        for metadata_modification_type in [models.MetadataModificationType.ADD, models.MetadataModificationType.REMOVE]:
            tx = models.ModifyAccountMetadataTransaction.create(
                deadline=models.Deadline.create(),
                network_type=config.network_type,
                metadata_type=models.MetadataType.ADDRESS,
                metadata_id=alice.address,
                modifications=[
                    models.MetadataModification(metadata_modification_type, models.Field(metadata_key, 'bar')),
                    models.MetadataModification(metadata_modification_type, models.Field(metadata_key + '2', 'bar')),
                ],
            )

            signed_tx = tx.sign_with(alice, config.gen_hash)

            tx = await announce(signed_tx)
            self.assertEqual(isinstance(tx, models.ModifyAccountMetadataTransaction), True)
            self.assertEqual(tx.metadata_id, alice.address)
            self.assertEqual(len(tx.modifications) > 0, True)
            self.assertEqual(tx.modifications[0].field.key, metadata_key)
            self.assertEqual(tx.modifications[0].field.value, 'bar')

    async def test_modify_mosaic_metadata_transaction(self):
        alice = self.t4.pop()

        mosaic_id = await helper.create_mosaic(alice, 1)
        metadata_key = 'foo' + hexlify(os.urandom(4)).decode('ascii')

        for metadata_modification_type in [models.MetadataModificationType.ADD, models.MetadataModificationType.REMOVE]:
            tx = models.ModifyMosaicMetadataTransaction.create(
                deadline=models.Deadline.create(),
                network_type=config.network_type,
                metadata_type=models.MetadataType.MOSAIC,
                metadata_id=mosaic_id,
                modifications=[
                    models.MetadataModification(metadata_modification_type, models.Field(metadata_key, 'bar')),
                    models.MetadataModification(metadata_modification_type, models.Field(metadata_key + '2', 'bar')),
                ],
            )

            signed_tx = tx.sign_with(alice, config.gen_hash)

            tx = await announce(signed_tx)
            self.assertEqual(isinstance(tx, models.ModifyMosaicMetadataTransaction), True)
            self.assertEqual(tx.metadata_id, mosaic_id)
            self.assertEqual(len(tx.modifications) > 0, True)
            self.assertEqual(tx.modifications[0].field.key, metadata_key)
            self.assertEqual(tx.modifications[0].field.value, 'bar')

    async def test_modify_namespace_metadata_transaction(self):
        alice = self.t5.pop()

        namespace = 'foo' + hexlify(os.urandom(4)).decode('utf-8') + ".bar"
        await helper.create_namespace(alice, namespace)

        metadata_key = 'foo' + hexlify(os.urandom(4)).decode('ascii')

        for metadata_modification_type in [models.MetadataModificationType.ADD, models.MetadataModificationType.REMOVE]:
            tx = models.ModifyNamespaceMetadataTransaction.create(
                deadline=models.Deadline.create(),
                network_type=config.network_type,
                metadata_type=models.MetadataType.NAMESPACE,
                metadata_id=models.NamespaceId(namespace),
                modifications=[
                    models.MetadataModification(metadata_modification_type, models.Field(metadata_key, 'bar')),
                    models.MetadataModification(metadata_modification_type, models.Field(metadata_key + '2', 'bar')),
                ],
            )

            signed_tx = tx.sign_with(alice, config.gen_hash)

            tx = await announce(signed_tx)
            self.assertEqual(isinstance(tx, models.ModifyNamespaceMetadataTransaction), True)
            self.assertEqual(tx.metadata_id, models.NamespaceId(namespace))
            self.assertEqual(len(tx.modifications) > 0, True)
            self.assertEqual(tx.modifications[0].field.key, metadata_key)
            self.assertEqual(tx.modifications[0].field.value, 'bar')

    async def test_account_metadata(self):
        alice = self.t2.pop()

        self.modifications = [
            models.MetadataModification(models.MetadataModificationType.ADD, models.Field('foo', 'bar')),
            models.MetadataModification(models.MetadataModificationType.ADD, models.Field('foo2', 'bar')),
        ]

        tx = models.ModifyAccountMetadataTransaction.create(
            deadline=models.Deadline.create(),
            network_type=config.network_type,
            metadata_type=models.MetadataType.ADDRESS,
            metadata_id=alice.address,
            modifications=self.modifications,
        )

        signed_tx = tx.sign_with(alice, config.gen_hash)

        tx = await announce(signed_tx)

        with client.MetadataHTTP(config.ENDPOINT, network_type=config.network_type) as http:
            reply = http.get_account_metadata(alice)
            self.assertEqual(isinstance(reply, models.AddressMetadataInfo), True)
            self.assertEqual(len(reply.metadata.flds), 2)
            self.assertEqual(reply.metadata.flds[0], self.modifications[0].field)
            self.assertEqual(reply.metadata.flds[1], self.modifications[1].field)

            reply = http.get_metadata(alice.address.plain())
            self.assertEqual(isinstance(reply, models.MetadataInfo), True)
            self.assertEqual(isinstance(reply.metadata, models.AddressMetadata), True)
            self.assertEqual(len(reply.metadata.flds), 2)
            self.assertEqual(reply.metadata.flds[0], self.modifications[0].field)
            self.assertEqual(reply.metadata.flds[1], self.modifications[1].field)

            reply = http.get_metadatas([alice.address.plain()])
            self.assertEqual(len(reply) > 0, True)
            self.assertEqual(isinstance(reply[0], models.MetadataInfo), True)
            self.assertEqual(isinstance(reply[0].metadata, models.AddressMetadata), True)
            self.assertEqual(len(reply[0].metadata.flds), 2)
            self.assertEqual(reply[0].metadata.flds[0], self.modifications[0].field)
            self.assertEqual(reply[0].metadata.flds[1], self.modifications[1].field)

    async def test_mosaic_metadata(self):
        alice = self.t4.pop()

        mosaic_id = await helper.create_mosaic(alice, 1)

        self.modifications = [
            models.MetadataModification(models.MetadataModificationType.ADD, models.Field('foo', 'bar')),
            models.MetadataModification(models.MetadataModificationType.ADD, models.Field('foo2', 'bar')),
        ]

        tx = models.ModifyMosaicMetadataTransaction.create(
            deadline=models.Deadline.create(),
            network_type=config.network_type,
            metadata_type=models.MetadataType.MOSAIC,
            metadata_id=mosaic_id,
            modifications=self.modifications,
        )

        signed_tx = tx.sign_with(alice, config.gen_hash)

        tx = await announce(signed_tx)

        with client.MetadataHTTP(config.ENDPOINT, network_type=config.network_type) as http:
            reply = http.get_mosaic_metadata(mosaic_id)
            self.assertEqual(isinstance(reply, models.MosaicMetadataInfo), True)
            self.assertEqual(len(reply.metadata.flds), 2)
            self.assertEqual(reply.metadata.flds[0], self.modifications[0].field)
            self.assertEqual(reply.metadata.flds[1], self.modifications[1].field)

            reply = http.get_metadata(mosaic_id.get_id())
            self.assertEqual(isinstance(reply, models.MetadataInfo), True)
            self.assertEqual(isinstance(reply.metadata, models.MosaicMetadata), True)
            self.assertEqual(len(reply.metadata.flds), 2)
            self.assertEqual(reply.metadata.flds[0], self.modifications[0].field)
            self.assertEqual(reply.metadata.flds[1], self.modifications[1].field)

            reply = http.get_metadatas([mosaic_id.get_id()])
            self.assertEqual(len(reply) > 0, True)
            self.assertEqual(isinstance(reply[0], models.MetadataInfo), True)
            self.assertEqual(isinstance(reply[0].metadata, models.MosaicMetadata), True)
            self.assertEqual(len(reply[0].metadata.flds), 2)
            self.assertEqual(reply[0].metadata.flds[0], self.modifications[0].field)
            self.assertEqual(reply[0].metadata.flds[1], self.modifications[1].field)

    async def test_namespace_metadata(self):
        alice = self.t5.pop()

        self.namespace = 'foo' + hexlify(os.urandom(4)).decode('utf-8') + ".bar"
        await helper.create_namespace(alice, self.namespace)

        self.modifications = [
            models.MetadataModification(models.MetadataModificationType.ADD, models.Field('foo', 'bar')),
            models.MetadataModification(models.MetadataModificationType.ADD, models.Field('foo2', 'bar')),
        ]

        tx = models.ModifyNamespaceMetadataTransaction.create(
            deadline=models.Deadline.create(),
            network_type=config.network_type,
            metadata_type=models.MetadataType.NAMESPACE,
            metadata_id=models.NamespaceId(self.namespace),
            modifications=self.modifications,
        )

        signed_tx = tx.sign_with(alice, config.gen_hash)

        tx = await announce(signed_tx)

        with client.MetadataHTTP(config.ENDPOINT, network_type=config.network_type) as http:
            reply = http.get_namespace_metadata(models.NamespaceId(self.namespace))
            self.assertEqual(isinstance(reply, models.NamespaceMetadataInfo), True)
            self.assertEqual(len(reply.metadata.flds), 2)
            self.assertEqual(reply.metadata.flds[0], self.modifications[0].field)
            self.assertEqual(reply.metadata.flds[1], self.modifications[1].field)

            reply = http.get_metadata(models.NamespaceId(self.namespace).get_id())
            self.assertEqual(isinstance(reply, models.MetadataInfo), True)
            self.assertEqual(isinstance(reply.metadata, models.NamespaceMetadata), True)
            self.assertEqual(len(reply.metadata.flds), 2)
            self.assertEqual(reply.metadata.flds[0], self.modifications[0].field)
            self.assertEqual(reply.metadata.flds[1], self.modifications[1].field)

            reply = http.get_metadatas([models.NamespaceId(self.namespace).get_id()])
            self.assertEqual(len(reply) > 0, True)
            self.assertEqual(isinstance(reply[0], models.MetadataInfo), True)
            self.assertEqual(isinstance(reply[0].metadata, models.NamespaceMetadata), True)
            self.assertEqual(len(reply[0].metadata.flds), 2)
            self.assertEqual(reply[0].metadata.flds[0], self.modifications[0].field)
            self.assertEqual(reply[0].metadata.flds[1], self.modifications[1].field)

    def test_get_mosaic(self):
        with client.MosaicHTTP(config.ENDPOINT, network_type=config.network_type) as http:
            info = http.get_mosaic(config.mosaic_id)
            self.assertEqual(isinstance(info, models.MosaicInfo), True)

    def test_get_mosaics(self):
        with client.MosaicHTTP(config.ENDPOINT, network_type=config.network_type) as http:
            info = http.get_mosaics([config.mosaic_id])
            self.assertEqual(len(info), 1)
            self.assertEqual(isinstance(info[0], models.MosaicInfo), True)

    def test_get_mosaic_names(self):
        with client.MosaicHTTP(config.ENDPOINT, network_type=config.network_type) as http:
            info = http.get_mosaic_names([config.mosaic_id])
            self.assertEqual(len(info), 1),
            self.assertEqual(isinstance(info[0], models.MosaicName), True)
            self.assertEqual(info[0].names[0], 'prx.xpx')

    def test_get_namespace(self):
        with client.NamespaceHTTP(config.ENDPOINT, network_type=config.network_type) as http:
            info = http.get_namespace(models.NamespaceId('prx.xpx'))
            self.assertEqual(isinstance(info, models.NamespaceInfo), True)
            self.assertEqual(info.owner.public_key, config.nemesis.public_key.upper())

    def test_get_namespaces_from_account(self):
        with client.NamespaceHTTP(config.ENDPOINT, network_type=config.network_type) as http:
            info = http.get_namespaces_from_account(config.nemesis.address)
            self.assertEqual(len(info) > 0, True)
            self.assertEqual(isinstance(info[0], models.NamespaceInfo), True)

    def test_get_namespaces_from_accounts(self):
        with client.NamespaceHTTP(config.ENDPOINT, network_type=config.network_type) as http:
            info = http.get_namespaces_from_accounts([config.nemesis.address])
            self.assertEqual(len(info) > 0, True)
            self.assertEqual(isinstance(info[0], models.NamespaceInfo), True)

    def test_get_namespaces_name(self):
        with client.NamespaceHTTP(config.ENDPOINT, network_type=config.network_type) as http:
            info = http.get_namespaces_name([models.NamespaceId('prx.xpx')])
            self.assertEqual(len(info), 1)
            self.assertEqual(info[0].name, 'prx.xpx')

    def test_get_network_type(self):
        with client.NetworkHTTP(config.ENDPOINT, network_type=config.network_type) as http:
            info = http.get_network_type()
            self.assertEqual(isinstance(info, models.NetworkType), True)

    def test_get_node_info(self):
        with client.NodeHTTP(config.ENDPOINT, network_type=config.network_type) as http:
            info = http.get_node_info()
            self.assertEqual(isinstance(info, models.NodeInfo), True)

    def test_get_node_time(self):
        with client.NodeHTTP(config.ENDPOINT, network_type=config.network_type) as http:
            info = http.get_node_time()
            self.assertEqual(isinstance(info, models.NodeTime), True)



"""
    account_meta
    ================

    Account metadata.

    License
    -------

    Copyright 2019 NEM

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""

from __future__ import annotations

from ..blockchain.network_type import OptionalNetworkType
from ... import util

__all__ = ['AccountMeta']


@util.inherit_doc
@util.dataclass(frozen=True)
class AccountMeta(util.DTO):
    """
    Meta describing an account.

    DTO Format:
        .. code-block:: yaml

            AccountMetaDTO: null
    """

    # height: int
    # hash: str
    # merkle_component_hash: str
    # index: int
    # id: str

    @classmethod
    def validate_dto(cls, data: dict) -> bool:
        """Validate the data-transfer object."""

        # required_l1 = {'height', 'hash', 'merkleComponentHash', 'index', 'id'}

        # return (
        # cls.validate_dto_required(data, required_l1)
        # and cls.validate_dto_all(data, required_l1)
        # )

        return (data == {})

    def to_dto(
        self,
        network_type: OptionalNetworkType = None,
    ) -> dict:
        # return {
        # 'height': util.u64_to_dto(self.height),
        # 'hash': self.hash,
        # 'merkleComponentHash': self.merkle_component_hash,
        # 'index': self.index,  # Swagger specify this as an 'integer' (no clue about size)
        # 'id': self.id
        # }

        return {}

    @classmethod
    def create_from_dto(
        cls,
        data: dict,
        network_type: OptionalNetworkType = None,
    ):
        if not cls.validate_dto(data):
            raise ValueError('Invalid data-transfer object.')

        # return cls(
            # height=util.u64_from_dto(data['height']),
            # hash=data['hash'],
            # merkle_component_hash=data['merkleComponentHash'],
            # index=data['index'],  # Swagger specify this as an 'integer' (no clue about size)
            # id=data['id'],
        # )

        return cls()



"""
    account_info
    ============

    Account data and metadata pair.

    License
    -------

    Copyright 2019 NEM

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""

from __future__ import annotations

from .account_meta import AccountMeta
from .address import Address
from .public_account import PublicAccount
from ..blockchain.network_type import OptionalNetworkType
from ..mosaic.mosaic import Mosaic, MosaicList
from ... import util

__all__ = ['AccountInfo']


@util.inherit_doc
@util.dataclass(frozen=True)
class AccountInfo(util.DTO):
    """
    Basic information describing an account.

    :param meta: Account metadata.
    :param address: Account address.
    :param address_height: Chain height when address was published.
    :param public_key: Account public key.
    :param public_key_height: Chain height when public key was published.
    :param mosaics: List of mosaics owned by account.

    DTO Format:
        .. code-block:: yaml

            AccountDTO:
                # Hex(Address) (50-bytes)
                address: string
                addressHeight: UInt64DTO
                # Hex(PublicKey) (64-bytes)
                publicKey: string
                publicKeyHeight: UInt64DTO
                accountType: UInt64DTO
                linkedAccountType: string
                mosaics: MosaicDTO[]

            AccountInfoDTO:
                meta: AccountMetaDTO
                account: AccountDTO
    """

    meta: AccountMeta
    address: Address
    address_height: int
    public_key: str
    public_key_height: int
    account_type: int
    linked_account_key: str
    mosaics: MosaicList
    snapshots: dict

    @property
    def public_account(self) -> PublicAccount:
        """Get public account."""
        return PublicAccount(self.address, self.public_key)

    @classmethod
    def validate_dto(cls, data: dict) -> bool:
        """Validate the data-transfer object."""

        required_l1 = {'meta', 'account'}
        required_l2 = {
            'address',
            'addressHeight',
            'publicKey',
            'publicKeyHeight',
            'accountType',
            'linkedAccountKey',
            'mosaics',
            'snapshots'
        }
        return (
            # Level 1
            cls.validate_dto_required(data, required_l1)
            and cls.validate_dto_all(data, required_l1)
            # Level 2
            and cls.validate_dto_required(data['account'], required_l2)
            and cls.validate_dto_all(data['account'], required_l2)
        )

    def to_dto(
        self,
        network_type: OptionalNetworkType = None,
    ) -> dict:
        meta = self.meta.to_dto(network_type)
        account = {
            'address': util.hexlify(self.address.encoded),
            'addressHeight': util.u64_to_dto(self.address_height),
            'publicKey': self.public_key,
            'publicKeyHeight': util.u64_to_dto(self.public_key_height),
            'accountType': self.account_type,
            'linkedAccountKey': self.linked_account_key,
            'mosaics': Mosaic.sequence_to_dto(self.mosaics, network_type),
            'snapshots': self.snapshots
        }

        return {
            'meta': meta,
            'account': account,
        }

    @classmethod
    def create_from_dto(
        cls,
        data: dict,
        network_type: OptionalNetworkType = None,
    ):
        if not cls.validate_dto(data):
            raise ValueError('Invalid data-transfer object.')

        meta = data['meta']
        account = data['account']
        return cls(
            meta=AccountMeta.create_from_dto(meta, network_type),
            address=Address.create_from_encoded(account['address']),
            address_height=util.u64_from_dto(account.get('addressHeight', [0, 0])),
            public_key=account['publicKey'],
            public_key_height=util.u64_from_dto(account.get('publicKeyHeight', [0, 0])),
            account_type=account['accountType'],
            linked_account_key=account['linkedAccountKey'],
            mosaics=Mosaic.sequence_from_dto(account.get('mosaics', []), network_type),
            snapshots=account.get('snapshots', [])
        )



import asyncio

from xpxchain import util
from tests import harness


class TestLoop(harness.TestCase):

    def test_get_event_loop(self):
        new_loop = asyncio.new_event_loop()
        self.assertIs(asyncio.get_event_loop(), util.get_event_loop())
        self.assertIs(new_loop, util.get_event_loop(new_loop))

        new_loop.close()

    def test_no_running_loop(self):
        with self.assertRaises(RuntimeError):
            util.get_running_loop()

    async def test_get_running_loop(self):
        self.assertIs(self.loop, util.get_running_loop())




from xpxchain import client
from xpxchain import models
from xpxchain import errors
from tests import config
import asyncio


async def announce(tx):
    async with client.Listener(f'{config.ENDPOINT}/ws', network_type=config.network_type) as listener:
        address = models.PublicAccount.create_from_public_key(tx.signer, config.network_type).address

        await listener.confirmed(address)
        await listener.status(address)

        async with client.AsyncTransactionHTTP(config.ENDPOINT, network_type=config.network_type) as http:
            await http.announce(tx)

        async for m in listener:
            if ((m.channel_name == 'status') and (m.message.hash == tx.hash.upper())):
                raise errors.TransactionError(m.message)
            elif ((m.channel_name == 'confirmedAdded') and (m.message.transaction_info.hash == tx.hash.upper())):
                return m.message


async def announce_partial(tx):
    async with client.Listener(f'{config.ENDPOINT}/ws', network_type=config.network_type) as listener:
        address = models.PublicAccount.create_from_public_key(tx.signer, config.network_type).address

        await listener.aggregate_bonded_added(address)
        await listener.status(address)

        async with client.AsyncTransactionHTTP(config.ENDPOINT, network_type=config.network_type) as http:
            await http.announce_partial(tx)

        async for m in listener:
            if ((m.channel_name == 'status') and (m.message.hash == tx.hash.upper())):
                raise errors.TransactionError(m.message)
            elif ((m.channel_name == 'partialAdded') and (m.message.transaction_info.hash == tx.hash.upper())):
                return m.message


async def announce_cosignature(tx):
    async with client.Listener(f'{config.ENDPOINT}/ws', network_type=config.network_type) as listener:
        address = models.PublicAccount.create_from_public_key(tx.signer, config.network_type).address

        await listener.confirmed(address)
        await listener.status(address)

        async with client.AsyncTransactionHTTP(config.ENDPOINT, network_type=config.network_type) as http:
            await http.announce_cosignature(tx)

        async for m in listener:
            if ((m.channel_name == 'status') and (m.message.hash == tx.hash.upper())):
                raise errors.TransactionError(m.message)
            elif ((m.channel_name == 'confirmedAdded') and (m.message.transaction_info.hash == tx.parent_hash.upper())):
                return m.message


async def send_funds(sender, recipient, amount, quiet=True):
    tx = models.TransferTransaction.create(
        deadline=models.Deadline.create(),
        recipient=recipient.address,
        mosaics=[models.Mosaic(config.mosaic_id, amount)],
        network_type=config.network_type,
    )

    signed_tx = tx.sign_with(sender, config.gen_hash)

    if (not quiet):
        print(f"Sending funds to {recipient.address.address} {signed_tx.hash}")

    return await announce(signed_tx)


async def prepare(plan, max_run=100):
    aws = plan[:max_run]
    plan = plan[max_run:]
    hashes = []

    while (aws):
        done, pending = await asyncio.wait(aws, return_when=asyncio.FIRST_COMPLETED)

        aws = []

        for task in done:
            try:
                tx = task.result()
            except errors.TransactionError as e:
                print(f"Error transaction {e.hash}: {e}")

            hashes.append(tx.transaction_info.hash)

        for task in pending:
            aws.append(task)

        planned = len(aws)
        if (planned < max_run):
            diff = max_run - planned
            aws += plan[:diff]
            plan = plan[diff:]

    return hashes


async def create_namespace(account, namespace):
    namespace_name = namespace.split(".")

    tx = models.RegisterNamespaceTransaction.create_root_namespace(
        deadline=models.Deadline.create(),
        network_type=config.network_type,
        namespace_name=namespace_name[0],
        duration=60
    )

    signed_tx = tx.sign_with(account, config.gen_hash)

    tx = await announce(signed_tx)
    n1 = models.NamespaceName(tx.namespace_id, tx.namespace_name)

    if (len(namespace_name) > 1):
        # Create sub namespace
        tx = models.RegisterNamespaceTransaction.create_sub_namespace(
            deadline=models.Deadline.create(),
            network_type=config.network_type,
            namespace_name=namespace_name[1],
            parent_namespace=namespace_name[0]
        )

        signed_tx = tx.sign_with(account, config.gen_hash)

        tx = await announce(signed_tx)
        n2 = models.NamespaceName(tx.namespace_id, tx.namespace_name, n1.namespace_id)

        return n1, n2

    return n1


async def create_mosaic(account, nonce):
    nonce = models.MosaicNonce(nonce)
    mosaic_id = models.MosaicId.create_from_nonce(nonce, account)

    tx = models.MosaicDefinitionTransaction.create(
        deadline=models.Deadline.create(),
        network_type=config.network_type,
        nonce=nonce,
        mosaic_id=mosaic_id,
        mosaic_properties=models.MosaicProperties(0x3, 3),
    )

    signed_tx = tx.sign_with(account, config.gen_hash)

    tx = await announce(signed_tx)

    return mosaic_id



"""
    account_property
    ================

    Property for an account.

    License
    -------

    Copyright 2019 NEM

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""

from __future__ import annotations
import typing

from .property_type import PropertyType
from .address import Address
from ..blockchain.network_type import OptionalNetworkType
from ..mosaic.mosaic_id import MosaicId
from ..transaction.transaction_type import TransactionType
from ... import util

__all__ = ['AccountProperty']

PropertyValue = typing.Union[
    Address,
    MosaicId,
    TransactionType,
]
PropertyValueList = typing.Sequence[PropertyValue]


def to_dto(
    property_type: int,
    values: typing.Sequence[PropertyValue],
) -> typing.List[str]:
    """Export properties to base64."""

    if property_type & 0x01:
        # We have an address.
        return [i.hex for i in values]  # type: ignore
    elif property_type & 0x02:
        # We have a mosaic.
        return [util.u64_to_dto(i) for i in values]  # type: ignore
    elif property_type & 0x04:
        # We have a transaction.
        return [i.to_dto() for i in values]  # type: ignore
    else:
        raise ValueError('Unknown property type.')


def from_dto(
    property_type: int,
    values: typing.List[str],
) -> typing.Sequence[PropertyValue]:
    """Load properties from base64."""

    if property_type & 0x01:
        # We have an address.
        return [Address.create_from_encoded(i) for i in values]  # type: ignore
    elif property_type & 0x02:
        # We have a mosaic.
        return [MosaicId(util.u64_from_dto(i)) for i in values]  # type: ignore
    elif property_type & 0x04:
        # We have a transaction.
        return [TransactionType.create_from_dto(i) for i in values]  # type: ignore
    else:
        raise ValueError('Unknown property type.')


@util.inherit_doc
@util.dataclass(frozen=True)
class AccountProperty(util.DTO):
    """
    Describe account property via type and values.

    :param property_type: Account property type.
    :param values: Property values.

    DTO Format:
        .. code-block:: yaml

            AccountPropertyDTO:
                propertyType: integer
                # Base64(Address), Base64(MosaicID), Base64(Transaction)
                values: string[]
    """

    property_type: PropertyType
    values: typing.Sequence[PropertyValue]

    @classmethod
    def validate_dto(cls, data: dict) -> bool:
        """Validate the data-transfer object."""

        required_keys = {'propertyType', 'values'}
        return (
            cls.validate_dto_required(data, required_keys)
            and cls.validate_dto_all(data, required_keys)
        )

    def to_dto(
        self,
        network_type: OptionalNetworkType = None,
    ) -> dict:
        return {
            'propertyType': self.property_type.to_dto(network_type),
            'values': to_dto(int(self.property_type), self.values),
        }

    @classmethod
    def create_from_dto(
        cls,
        data: dict,
        network_type: OptionalNetworkType = None,
    ):
        if not cls.validate_dto(data):
            raise ValueError('Invalid data-transfer object.')

        property_type = data['propertyType']
        values = data['values']
        return cls(
            property_type=PropertyType.create_from_dto(property_type, network_type),
            values=from_dto(property_type, values),
        )



"""
    message
    =======

    Abstract message class as base for message types.

    License
    -------

    Copyright 2019 NEM

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""

from __future__ import annotations
import typing

from .message_type import MessageType
from ..blockchain.network_type import OptionalNetworkType
from ... import util

__all__ = ['Message']


@util.dataclass(frozen=True)
class Message(util.Model):
    """
    Abstract message type.

    :param type: Message type.
    :param payload: Message data, in bytes.
    """

    type: MessageType
    payload: bytes

    @classmethod
    def create(cls, data: typing.AnyStr):
        """Create a message from raw bytes."""
        raise util.AbstractMethodError

    @classmethod
    def validate_dto(cls, data: dict) -> bool:
        """Validate the data-transfer object."""

        required_keys = {'type', 'payload'}
        return (
            cls.validate_dto_required(data, required_keys)
            and cls.validate_dto_all(data, required_keys)
        )

    def to_dto(
        self,
        network_type: OptionalNetworkType = None,
    ) -> dict:
        return {
            'type': self.type.to_dto(network_type),
            'payload': util.hexlify(self.payload),
        }

    @classmethod
    def create_from_dto(
        cls,
        data: dict,
        network_type: OptionalNetworkType = None,
    ):
        if not cls.validate_dto(data):
            raise ValueError('Invalid data-transfer object.')

        payload = util.unhexlify(data['payload'])
        return cls.create(payload)

    def to_catbuffer(
        self,
        network_type: OptionalNetworkType = None,
        fee_strategy: util.FeeCalculationStrategy = util.FeeCalculationStrategy.MEDIUM,
    ) -> bytes:
        return util.u8_to_catbuffer(self.type) + self.payload

    @classmethod
    def create_from_catbuffer_pair(
        cls,
        data: bytes,
        network_type: OptionalNetworkType = None,
    ):
        return cls.create(data[1:]), data[len(data):]

    def catbuffer_size_specific(self) -> int:
        return util.U8_BYTES + len(self.payload)



"""
    aggregate_transaction_info
    ==========================

    Aggregate transaction metadata.

    License
    -------

    Copyright 2019 NEM

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""

from __future__ import annotations

from ..blockchain.network_type import OptionalNetworkType
from ... import util

__all__ = ['AggregateTransactionInfo']


@util.inherit_doc
@util.dataclass(frozen=True)
class AggregateTransactionInfo(util.DTO):
    """Aggregate transaction metadata."""

    height: int
    index: int
    id: str
    aggregate_hash: str
    aggregate_id: str

    def is_unconfirmed(self) -> bool:
        """Is transaction pending to be included."""

        return (
            self.height == 0
            and self.aggregate_hash == self.aggregate_id
        )

    def is_confirmed(self) -> bool:
        """Is transaction already included."""
        return self.height > 0

    def has_missing_signatures(self) -> bool:
        """Does the transaction have missing signatures."""
        return (
            self.height == 0
            and self.aggregate_hash != self.aggregate_id
        )

    @classmethod
    def validate_dto(cls, data: dict) -> bool:
        """Validate the data-transfer object."""

        required_keys = {
            'height',
            'index',
            'id',
            'aggregateHash',
            'aggregateId',
        }
        return (
            cls.validate_dto_required(data, required_keys)
            and cls.validate_dto_all(data, required_keys)
        )

    def to_dto(
        self,
        network_type: OptionalNetworkType = None,
    ) -> dict:
        return {
            'height': util.u64_to_dto(self.height),
            'index': self.index,
            'id': self.id,
            'aggregateHash': self.aggregate_hash,
            'aggregateId': self.aggregate_id,
        }

    @classmethod
    def create_from_dto(
        cls,
        data: dict,
        network_type: OptionalNetworkType = None,
    ):
        if not cls.validate_dto(data):
            raise ValueError('Invalid data-transfer object.')

        return cls(
            height=util.u64_from_dto(data['height']),
            index=data['index'],
            id=data['id'],
            aggregate_hash=data['aggregateHash'],
            aggregate_id=data['aggregateId'],
        )



import warnings

# We want to catch all non-user warnings, which are emitted
# from the fallback ED25519 implementation.
warnings.filterwarnings('error')
warnings.filterwarnings('default', category=UserWarning)



"""
    resolution_statement
    ====================

    Resolution statement.

    License
    -------

    Copyright 2019 NEM

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""

from __future__ import annotations
import typing

from .resolution_entry import ResolutionEntry
from ..blockchain.network_type import OptionalNetworkType
from ... import util

__all__ = [
    'ResolutionStatement',
]


@util.inherit_doc
@util.dataclass(frozen=True)
class ResolutionStatement(util.DTO):
    """
    A resolution statement keeps the relation between a namespace alias used in a transaction
    and the real address or mosaicId.

    :param height:
    :param unresolved:
    :param resolutionEntries: The array of resolution entries linked to the unresolved namespaceId.
    """

    height: int
    unresolved: int
    resolution_entries: typing.Sequence[ResolutionEntry]

    # DTO

    @classmethod
    def validate_dto(cls, data: dict) -> bool:
        required_keys = {'height', 'unresolved', 'resolutionEntries'}
        return (
            cls.validate_dto_required(data, required_keys)
            and cls.validate_dto_all(data, required_keys)
        )

    def to_dto(
        self,
        network_type: OptionalNetworkType = None,
    ) -> dict:
        return {
            'height': util.u64_to_dto(self.height),
            'unresolved': util.u64_to_dto(self.unresolved),
            'resolutionEntries': [ResolutionEntry.to_dto(i) for i in self.resolution_entries]
        }

    @classmethod
    def create_from_dto(
        cls,
        data: dict,
        network_type: OptionalNetworkType = None,
    ):
        if not cls.validate_dto(data):
            raise ValueError('Invalid data-transfer object.')

        return cls(
            height=util.u64_from_dto(data['mosaicId']),
            unresolved=util.u64_from_dto(data['unresolved']),
            resolution_entries=[ResolutionEntry.create_from_dto(i) for i in data['resolutionEntries']]
        )



"""
    client
    ======

    HTTP client wrapper to inject the proper host (schema, domain and port)
    into requests providing solely a relative path.
"""

from __future__ import annotations
import typing
import urllib.error
import urllib3

from .. import util

# UTILITY


DEFAULT_PORT = {
    'http': 80,
    'https': 443,
    'ws': 80,
    'wss': 443
}


def parse_http_url(uri: str, scheme: str = 'http') -> urllib3.util.Url:
    """Add a default scheme if not present."""

    url = urllib3.util.parse_url(uri)
    if url.scheme is None:
        url = url._replace(scheme=scheme)
    if url.scheme not in ('http', 'https'):
        raise urllib.error.URLError("HTTP scheme not recognized.")
    if url.port is None:
        url._replace(port=DEFAULT_PORT[url.scheme])

    return url


def parse_ws_url(uri: str, scheme: str = 'ws') -> urllib3.util.Url:
    """Process a websockets URI and return the kwds for the initializer."""

    url = urllib3.util.parse_url(uri)
    if url.scheme is None:
        url = url._replace(scheme=scheme)
    if url.scheme not in ('ws', 'wss'):
        raise urllib.error.URLError("websockets scheme not recognized.")
    if url.port is None:
        url._replace(port=DEFAULT_PORT[url.scheme])

    return url


# HTTP


class ClientSharedBase(util.Object):
    """Shared, abstract base class for sync and async HTTP clients."""

    _session: typing.Any
    _endpoint: str

    def __init__(self, session, endpoint) -> None:
        self._session = session
        self._endpoint = parse_http_url(endpoint).url

    def close(self):
        """Close the client session."""
        raise util.AbstractMethodError

    @property
    def closed(self) -> bool:
        """Get if client session has been closed."""
        raise util.AbstractMethodError

    def delete(self, relative_path, *args, **kwds):
        """
        Make DELETE request from relative path.

        :param relative_path: Relative path from endpoint prefixed with "/".
        :param \\*args: Optional positional arguments for request.
        :param \\**kwds: Optional keyword arguments for request.
        """

        path = self._endpoint + relative_path
        return self._session.delete(path, *args, **kwds)

    def get(self, relative_path, *args, **kwds):
        """
        Make GET request from relative path.

        :param relative_path: Relative path from endpoint prefixed with "/".
        :param \\*args: Optional positional arguments for request.
        :param \\**kwds: Optional keyword arguments for request.
        """

        path = self._endpoint + relative_path
        return self._session.get(path, *args, **kwds)

    def head(self, relative_path, *args, **kwds):
        """
        Make HEAD request from relative path.

        :param relative_path: Relative path from endpoint prefixed with "/".
        :param \\*args: Optional positional arguments for request.
        :param \\**kwds: Optional keyword arguments for request.
        """

        path = self._endpoint + relative_path
        return self._session.head(path, *args, **kwds)

    def options(self, relative_path, *args, **kwds):
        """
        Make OPTIONS request from relative path.

        :param relative_path: Relative path from endpoint prefixed with "/".
        :param \\*args: Optional positional arguments for request.
        :param \\**kwds: Optional keyword arguments for request.
        """

        path = self._endpoint + relative_path
        return self._session.options(path, *args, **kwds)

    def patch(self, relative_path, *args, **kwds):
        """
        Make PATCH request from relative path.

        :param relative_path: Relative path from endpoint prefixed with "/".
        :param \\*args: Optional positional arguments for request.
        :param \\**kwds: Optional keyword arguments for request.
        """

        path = self._endpoint + relative_path
        return self._session.patch(path, *args, **kwds)

    def post(self, relative_path, *args, **kwds):
        """
        Make POST request from relative path.

        :param relative_path: Relative path from endpoint prefixed with "/".
        :param \\*args: Optional positional arguments for request.
        :param \\**kwds: Optional keyword arguments for request.
        """

        path = self._endpoint + relative_path
        return self._session.post(path, *args, **kwds)

    def put(self, relative_path, *args, **kwds):
        """
        Make PUT request from relative path.

        :param relative_path: Relative path from endpoint prefixed with "/".
        :param \\*args: Optional positional arguments for request.
        :param \\**kwds: Optional keyword arguments for request.
        """

        path = self._endpoint + relative_path
        return self._session.put(path, *args, **kwds)


@util.inherit_doc
class Client(ClientSharedBase):
    """
    Client wrapper for an abstract HTTP session.

    :param session: Requests or aiohttp-like HTTP client session.
    :param endpoint: Domain name and port for the endpoint.
    """

    _closed: bool

    def __init__(self, session, endpoint) -> None:
        super().__init__(session, endpoint)
        self._closed = False

    def __enter__(self) -> Client:
        return self

    def __exit__(self, exc_type, exc, tb) -> None:
        self.close()

    def close(self) -> None:
        self._closed = True
        self._session.close()

    @property
    def closed(self) -> bool:
        return self._closed


@util.inherit_doc
class AsyncClient(ClientSharedBase):
    """
    Asynchronous client with a managed loop.

    :param session: Requests or aiohttp-like HTTP client session.
    :param endpoint: Domain name and port for the endpoint.
    :param loop: Event loop.
    """

    def __init__(self, session, endpoint) -> None:
        super().__init__(session, endpoint)

    def __enter__(self) -> AsyncClient:
        raise TypeError("Only use async with.")

    def __exit__(self, exc_type, exc, tb) -> None:
        pass

    async def __aenter__(self) -> AsyncClient:
        return self

    async def __aexit__(self) -> None:
        await self.close()

    async def close(self) -> None:
        await self._session.close()

    @property
    def closed(self) -> bool:
        return typing.cast(bool, self._session.closed)


# WEBSOCKETS


class WebsocketClient(util.Object):
    """Asynchronous host using websockets."""

    _session: typing.Any

    def __init__(self, session) -> None:
        self._session = session

    async def __aiter__(self) -> typing.AsyncIterator[typing.AnyStr]:
        async for message in self._session:
            yield typing.cast(typing.AnyStr, message)

    @property
    def closed(self) -> bool:
        """Get if client session has been closed."""
        result = self._session.closed
        return typing.cast(bool, result)

    async def recv(self) -> typing.AnyStr:
        """Receive next message."""
        result = await self._session.recv()
        return typing.cast(typing.AnyStr, result)

    async def send(self, data: typing.AnyStr) -> None:
        """Send message."""
        await self._session.send(data)

    async def ping(
        self,
        data: typing.Optional[bytes] = None
    ) -> typing.Awaitable[None]:
        """Send websocket a ping."""
        result = await self._session.ping(data)
        return typing.cast(typing.Awaitable[None], result)

    async def pong(self, data: bytes = b'') -> None:
        """Send websocket a pong."""
        await self._session.pong(data)



"""
    account_property_modification
    =============================

    Account property modification type and value.

    License
    -------

    Copyright 2019 NEM

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""

from __future__ import annotations

from .field import Field
from .metadata_modification_type import MetadataModificationType
from ..blockchain.network_type import NetworkType, OptionalNetworkType
from ... import util

__all__ = ['MetadataModification']


@util.inherit_doc
@util.dataclass(frozen=True)
class MetadataModification(util.DTO):
    """
    Metadata modification type and value.

    :param modification_type: Metadata modification type.
    :param field: Modification field.
    """

    modification_type: MetadataModificationType
    field: Field

    @classmethod
    def validate_dto(cls, data: dict) -> bool:
        """Validate the data-transfer object."""

        required_keys = {'modificationType', 'key', 'value'}
        return (
            cls.validate_dto_required(data, required_keys)
            and cls.validate_dto_all(data, required_keys)
        )

    def catbuffer_size_specific(self) -> int:
        modification_size = util.U32_BYTES
        modification_type_size = MetadataModificationType.CATBUFFER_SIZE
        key_size = util.U8_BYTES
        value_size = util.U16_BYTES
        field_size = self.field.catbuffer_size_specific()

        return modification_size + modification_type_size + key_size + value_size + field_size

    def to_catbuffer_specific(
        self,
        network_type: NetworkType,
    ) -> bytes:

        modification_size = util.u32_to_catbuffer(self.catbuffer_size_specific())
        modification_type = util.u8_to_catbuffer(self.modification_type)
        key_size = util.u8_to_catbuffer(len(self.field.key))
        value_size = util.u16_to_catbuffer(len(self.field.value))
        field = self.field.to_catbuffer_specific(network_type)

        return modification_size + modification_type + key_size + value_size + field

    def to_dto(
        self,
        network_type: OptionalNetworkType = None,
    ) -> dict:
        return {
            'modificationType': self.modification_type.to_dto(network_type),
            'key': self.field.key,
            'value': self.field.value,
        }

    @classmethod
    def create_from_dto(
        cls,
        data: dict,
        network_type: OptionalNetworkType = None,
    ):
        if not cls.validate_dto(data):
            raise ValueError('Invalid data-transfer object.')

        return cls(
            modification_type=MetadataModificationType.create_from_dto(
                data['modificationType']
            ),
            field=Field(data['key'], data['value']),
        )



"""
    to_async_generator
    ==================

    RxPy extension to convert an observable to an async iterator.

    License
    -------

    Copyright 2019 NEM

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""

from __future__ import annotations
import asyncio
import collections
import rx
from rx.internal import extensionmethod


class AsyncIterable:
    """Adapter class to convert an observable to an asynchronous iterator."""

    def __init__(self, observable: rx.Observable):
        self.observable = observable
        self.notifications: collections.deque = collections.deque()
        self.future: asyncio.Future = asyncio.Future()
        self.observable.materialize().subscribe(self.on_next)

    def __aiter__(self):
        return self

    async def __anext__(self):
        self.process()
        value = await self.future
        self.future = asyncio.Future()
        return value

    def process(self):
        if not self.notifications or self.future.done():
            return

        notification = self.notifications.popleft()
        if notification.kind == 'N':
            self.future.set_result(notification.value)
        elif notification.kind == 'E':
            self.future.set_exception(notification.exception)
        else:
            self.future.set_exception(StopAsyncIteration)

    def on_next(self, notification):
        self.notifications.append(notification)
        self.process()


def __aiter__(self):
    return self


async def __anext__(self):
    # Use a name that's unlikely to be ever touched.
    if not hasattr(self, '_xpxchain_async_iter'):
        self._xpxchain_async_iter = AsyncIterable(self)
    try:
        return await self._xpxchain_async_iter.__anext__()
    except StopAsyncIteration:
        self._xpxchain_async_iter = None
        raise


extensionmethod(rx.AnonymousObservable)(__aiter__)
extensionmethod(rx.Observable)(__aiter__)
extensionmethod(rx.AnonymousObservable)(__anext__)
extensionmethod(rx.Observable)(__anext__)



from xpxchain import client
from xpxchain import models
from tests import harness
from tests import responses
from xpxchain import util


@harness.mocked_http_test_case({
    'clients': (client.MetadataHTTP, client.AsyncMetadataHTTP),
    'network_type': models.NetworkType.MIJIN_TEST,
    'tests': [
        {
            'name': 'test_get_account_metadata',
            'response': responses.ACCOUNT_METADATA["Ok"],
            'params': [models.PublicAccount.create_from_public_key('c55b4dfede7058a354aba9cff88dd6ba2d2006ab580bde849fb9d63481b462ab', models.NetworkType.MIJIN_TEST)],
            'method': 'get_account_metadata',
            'validation': [
                lambda x: (isinstance(x, models.AddressMetadataInfo), True),
                lambda x: (isinstance(x.metadata, models.AddressMetadata), True),
                lambda x: (x.metadata.metadata_type, models.MetadataType.ADDRESS),
                lambda x: (isinstance(x.metadata.metadata_id, models.Address), True),
                lambda x: (x.metadata.metadata_id.hex, '90ABBF4346E893A8EA4E420C6DBB9EBE9BE9180CD4056EAA28'),
                lambda x: (len(x.metadata.flds), 2),
                lambda x: (isinstance(x.metadata.flds[0], models.Field), True),
                lambda x: (x.metadata.flds[0].key, 'foo'),
                lambda x: (x.metadata.flds[0].value, 'bar'),
                lambda x: (x.metadata.flds[1].key, 'foo2'),
                lambda x: (x.metadata.flds[1].value, 'bar'),
            ],
        },
        {
            'name': 'test_get_mosaic_metadata',
            'response': responses.MOSAIC_METADATA["Ok"],
            'params': [models.MosaicId(util.u64_from_dto([1045581869, 105004269]))],
            'method': 'get_mosaic_metadata',
            'validation': [
                lambda x: (isinstance(x, models.MosaicMetadataInfo), True),
                lambda x: (isinstance(x.metadata, models.MosaicMetadata), True),
                lambda x: (x.metadata.metadata_type, models.MetadataType.MOSAIC),
                lambda x: (isinstance(x.metadata.metadata_id, models.MosaicId), True),
                lambda x: (x.metadata.metadata_id, models.MosaicId(util.u64_from_dto([1045581869, 105004269]))),
                lambda x: (len(x.metadata.flds), 2),
                lambda x: (isinstance(x.metadata.flds[0], models.Field), True),
                lambda x: (x.metadata.flds[0].key, 'foo'),
                lambda x: (x.metadata.flds[0].value, 'bar'),
                lambda x: (x.metadata.flds[1].key, 'foo2'),
                lambda x: (x.metadata.flds[1].value, 'bar'),
            ],
        },
        {
            'name': 'test_get_namespace_metadata',
            'response': responses.NAMESPACE_METADATA["Ok"],
            'params': [models.NamespaceId(util.u64_from_dto([130473079, 2260557970]))],
            'method': 'get_namespace_metadata',
            'validation': [
                lambda x: (isinstance(x, models.NamespaceMetadataInfo), True),
                lambda x: (isinstance(x.metadata, models.NamespaceMetadata), True),
                lambda x: (x.metadata.metadata_type, models.MetadataType.NAMESPACE),
                lambda x: (isinstance(x.metadata.metadata_id, models.NamespaceId), True),
                lambda x: (x.metadata.metadata_id, models.NamespaceId(util.u64_from_dto([130473079, 2260557970]))),
                lambda x: (len(x.metadata.flds), 2),
                lambda x: (isinstance(x.metadata.flds[0], models.Field), True),
                lambda x: (x.metadata.flds[0].key, 'foo'),
                lambda x: (x.metadata.flds[0].value, 'bar'),
                lambda x: (x.metadata.flds[1].key, 'foo2'),
                lambda x: (x.metadata.flds[1].value, 'bar'),
            ],
        },
        {
            'name': 'test_get_metadata',
            'response': responses.METADATA["Ok"],
            'params': [models.Address('SCV36Q2G5CJ2R2SOIIGG3O46X2N6SGAM2QCW5KRI')],
            'method': 'get_metadata',
            'validation': [
                lambda x: (isinstance(x, models.MetadataInfo), True),
                lambda x: (isinstance(x.metadata, models.AddressMetadata), True),
                lambda x: (x.metadata.metadata_type, models.MetadataType.ADDRESS),
                lambda x: (isinstance(x.metadata.metadata_id, models.Address), True),
                lambda x: (x.metadata.metadata_id.hex, '90ABBF4346E893A8EA4E420C6DBB9EBE9BE9180CD4056EAA28'),
                lambda x: (len(x.metadata.flds), 2),
                lambda x: (isinstance(x.metadata.flds[0], models.Field), True),
                lambda x: (x.metadata.flds[0].key, 'foo'),
                lambda x: (x.metadata.flds[0].value, 'bar'),
                lambda x: (x.metadata.flds[1].key, 'foo2'),
                lambda x: (x.metadata.flds[1].value, 'bar'),
            ],
        },
        {
            'name': 'test_get_metadatas',
            'response': responses.METADATAS["Ok"],
            'params': [[models.Address('SCV36Q2G5CJ2R2SOIIGG3O46X2N6SGAM2QCW5KRI')]],
            'method': 'get_metadatas',
            'validation': [
                lambda x: (len(x), 1),
                lambda x: (isinstance(x[0], models.MetadataInfo), True),
                lambda x: (isinstance(x[0].metadata, models.AddressMetadata), True),
                lambda x: (x[0].metadata.metadata_type, models.MetadataType.ADDRESS),
                lambda x: (isinstance(x[0].metadata.metadata_id, models.Address), True),
                lambda x: (x[0].metadata.metadata_id.hex, '90ABBF4346E893A8EA4E420C6DBB9EBE9BE9180CD4056EAA28'),
                lambda x: (len(x[0].metadata.flds), 2),
                lambda x: (isinstance(x[0].metadata.flds[0], models.Field), True),
                lambda x: (x[0].metadata.flds[0].key, 'foo'),
                lambda x: (x[0].metadata.flds[0].value, 'bar'),
                lambda x: (x[0].metadata.flds[1].key, 'foo2'),
                lambda x: (x[0].metadata.flds[1].value, 'bar'),
            ],
        },
    ],
})
class TestMetadataHTTP(harness.TestCase):
    pass



from xpxchain import client
from xpxchain import models
from tests import harness
from tests import responses


@harness.mocked_http_test_case({
    'clients': (client.NetworkHTTP, client.AsyncNetworkHTTP),
    'network_type': models.NetworkType.MIJIN_TEST,
    'tests': [
        {
            'name': 'test_get_network_type',
            'response': responses.NETWORK_TYPE['MIJIN_TEST'],
            'params': [],
            'method': 'get_network_type',
            'validation': [
                lambda x: (x, models.NetworkType.MIJIN_TEST),
            ]
        },
    ],
})
class TestNetworkHTTP(harness.TestCase):
    pass



"""
    modify_account_property_entity_type_transaction
    ===============================================

    Transaction to modify account property entity types.

    License
    -------

    Copyright 2019 NEM

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""

from __future__ import annotations
import typing

from .account_property_modification import AccountPropertyModification
from .deadline import Deadline
from .inner_transaction import InnerTransaction
from .registry import register_transaction
from .transaction import Transaction
from .transaction_info import TransactionInfo
from .transaction_type import TransactionType
from .transaction_version import TransactionVersion
from ..account.property_modification_type import PropertyModificationType
from ..account.property_type import PropertyType
from ..account.public_account import PublicAccount
from ..blockchain.network_type import NetworkType
from ... import util

__all__ = [
    'ModifyAccountPropertyEntityTypeTransaction',
    'ModifyAccountPropertyEntityTypeInnerTransaction',
]

AccountPropertyModificationList = typing.Sequence[AccountPropertyModification]
PROPERTY_TYPES = (
    PropertyType.ALLOW_TRANSACTION,
    PropertyType.BLOCK_TRANSACTION
)


def to_catbuffer(modification: AccountPropertyModification) -> bytes:
    """Serialize account property mosaic modification to catbuffer."""

    # uint8_t modification_type
    # uint16_t value
    type = modification.modification_type.to_catbuffer()
    value = util.u16_to_catbuffer(int(typing.cast(TransactionType, modification.value)))
    return type + value


def from_catbuffer(data: bytes) -> AccountPropertyModification:
    """Deserialize account property mosaic modification from catbuffer."""

    # uint8_t modification_type
    # uint16_t value
    modification_type = PropertyModificationType.create_from_catbuffer(data[:1])
    value = TransactionType(util.u16_from_catbuffer(data[1:3]))
    return AccountPropertyModification(modification_type, value)


@util.inherit_doc
@util.dataclass(frozen=True)
@register_transaction('MODIFY_ACCOUNT_PROPERTY_ENTITY_TYPE')
class ModifyAccountPropertyEntityTypeTransaction(Transaction):
    """
    Modify account property entity types transaction.

    :param network_type: Network type.
    :param version: Transaction version.
    :param deadline: Deadline to include transaction.
    :param max_fee: Max fee for the transaction. Higher fees increase priority.
    :param property_type: Account property type.
    :param modifications: List of account modifications.
    :param signature: (Optional) Transaction signature (missing if embedded transaction).
    :param signer: (Optional) Account of transaction creator.
    :param transaction_info: (Optional) Transaction metadata.
    """

    property_type: PropertyType
    modifications: AccountPropertyModificationList

    def __init__(
        self,
        network_type: NetworkType,
        version: TransactionVersion,
        deadline: Deadline,
        property_type: PropertyType,
        modifications: AccountPropertyModificationList,
        max_fee: int = 0,
        signature: typing.Optional[str] = None,
        signer: typing.Optional[PublicAccount] = None,
        transaction_info: typing.Optional[TransactionInfo] = None,
    ) -> None:
        if property_type not in PROPERTY_TYPES:
            raise ValueError('Property type must be a mosaic type.')
        if not all(i.is_entity_type() for i in modifications):
            raise ValueError('Modify account property entity types got invalid value.')
        super().__init__(
            TransactionType.MODIFY_ACCOUNT_PROPERTY_ENTITY_TYPE,
            network_type,
            version,
            deadline,
            max_fee,
            signature,
            signer,
            transaction_info,
        )
        self._set('property_type', property_type)
        self._set('modifications', modifications)

    @classmethod
    def create(
        cls,
        deadline: Deadline,
        property_type: PropertyType,
        modifications: AccountPropertyModificationList,
        network_type: NetworkType,
        max_fee: int = 0,
    ):
        """
        Create new modify account entity types transaction.

        :param deadline: Deadline to include transaction.
        :param property_type: Account property type.
        :param modifications: List of account modifications.
        :param network_type: Network type.
        :param max_fee: (Optional) Max fee defined by sender.
        """
        return cls(
            network_type,
            TransactionVersion.MODIFY_ACCOUNT_PROPERTY_ENTITY_TYPE,
            deadline,
            property_type,
            modifications,
            max_fee,
        )

    # CATBUFFER

    def catbuffer_size_specific(self) -> int:
        property_type_size = PropertyType.CATBUFFER_SIZE
        count_size = util.U8_BYTES
        modification_size = util.U8_BYTES + util.U16_BYTES
        modifications_size = modification_size * len(self.modifications)
        return property_type_size + count_size + modifications_size

    def to_catbuffer_specific(
        self,
        network_type: NetworkType,
    ) -> bytes:
        """Export modify account entity types-specific data to catbuffer."""

        # uint8_t property_type
        # uint8_t modifications_count
        # typedef Modification { type: uint8_t, value: uint16_t }
        # Modification[modifications_count] modifications
        property_type = self.property_type.to_catbuffer(network_type)
        modifications_count = util.u8_to_catbuffer(len(self.modifications))
        modifications = b''.join(map(to_catbuffer, self.modifications))
        return property_type + modifications_count + modifications

    def load_catbuffer_specific(
        self,
        data: bytes,
        network_type: NetworkType,
    ) -> bytes:
        """Load modify account entity types-specific data data from catbuffer."""

        # uint8_t property_type
        # uint8_t modifications_count
        # typedef Modification { type: uint8_t, value: uint16_t }
        # Modification[modifications_count] modifications
        property_type = PropertyType.create_from_catbuffer(data[:1], network_type)
        modifications_count = util.u8_from_catbuffer(data[1:2])
        data = data[2:]
        modifications = []
        for _ in range(modifications_count):
            modifications.append(from_catbuffer(data))
            data = data[3:]

        self._set('property_type', property_type)
        self._set('modifications', modifications)

        return data

    # DTO

    @classmethod
    def validate_dto_specific(cls, data: dict) -> bool:
        required_keys = {'propertyType', 'modifications'}
        return cls.validate_dto_required(data, required_keys)

    def to_dto_specific(
        self,
        network_type: NetworkType,
    ) -> dict:
        to_dto = AccountPropertyModification.sequence_to_dto
        return {
            'propertyType': self.property_type.to_dto(network_type),
            'modifications': to_dto(self.modifications, network_type),
        }

    def load_dto_specific(
        self,
        data: dict,
        network_type: NetworkType,
    ) -> None:
        from_dto = AccountPropertyModification.sequence_from_dto
        property_type = PropertyType.create_from_dto(data['propertyType'])
        modifications = from_dto(data['modifications'])
        self._set('property_type', property_type)
        self._set('modifications', modifications)


@register_transaction('MODIFY_ACCOUNT_PROPERTY_ENTITY_TYPE')
class ModifyAccountPropertyEntityTypeInnerTransaction(
    InnerTransaction,
    ModifyAccountPropertyEntityTypeTransaction
):
    """Embedded modify account entity types transaction."""

    __slots__ = ()



"""
    namespace_id
    ============

    Identifier for a namespace.

    License
    -------

    Copyright 2019 NEM

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""

import typing
from ... import util

__all__ = ['NamespaceId']

IdType = typing.Union[str, int]


def id_to_int(id: IdType) -> int:
    """Convert identifier to int."""

    if isinstance(id, int):
        return id
    elif isinstance(id, str):
        ids = util.generate_namespace_id(id) or [0]
        return ids[-1]
    else:
        raise TypeError(f"Invalid type for ID, got {type(id).__name__}")


@util.inherit_doc
@util.dataclass(frozen=True)
class NamespaceId(util.IntMixin, util.Object):
    """
    Identifier for a namespace.

    :param id: Identifier or name for namespace.
    """

    id: int

    def __init__(self, id: IdType) -> None:
        self._set('id', id_to_int(id))

    def __int__(self) -> int:
        return self.id

    def get_id(self) -> str:
        return util.hexlify(self.id.to_bytes(8, 'big'))

    @property
    def encoded(self) -> str:
        """Get the namespace ID as a hex-encoded string."""
        decoded = util.u64_to_catbuffer(self.id)
        return util.hexlify(decoded)

    @classmethod
    def create_from_encoded(cls, encoded: typing.AnyStr):
        """
        Create namespace ID from hex-encoded string.

        :param encoded: ID bytes or hex-encoded bytes.
        """
        decoded = util.decode_hex(encoded, with_prefix=True)
        return cls(util.u64_from_catbuffer(decoded))


NamespaceIdList = typing.Sequence[NamespaceId]



"""
    dto
    ===

    High-level NEM models.

    License
    -------

    Copyright 2019 NEM

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""

# Do not import any of these models into the in the
# __init__ of the subdirectories, since there is a
# complicated web of inter-dependencies within models.
# Just use glob imports at the models level.

# Account
from .account.account import *
from .account.account_info import *
from .account.account_meta import *
from .account.account_property import *
from .account.account_properties import *
from .account.address import *
from .account.multisig_account_graph_info import *
from .account.multisig_account_info import *
from .account.property_modification_type import *
from .account.property_type import *
from .account.public_account import *
from .account.account_names import *
from .account.account_balance import *
from .account import *

# Blockchain
from .blockchain.block_info import *
from .blockchain.block_type import *
from .blockchain.blockchain_score import *
from .blockchain.blockchain_storage_info import *
from .blockchain.blockchain_server_info import *
from .blockchain.merkle_path_item import *
from .blockchain.merkle_proof_info import *
from .blockchain.network_type import *
from .blockchain import *

# Receipt
from .receipt.balance_change_receipt import *
from .receipt.balance_transfer_receipt import *
from .receipt.artifact_expiry_receipt import *
from .receipt.inflation_receipt import *
from .receipt.receipt_base import *
from .receipt.receipt import *
from .receipt.receipt_type import *
from .receipt.receipt_version import *
from .receipt.source import *
from .receipt.statements import *
from .receipt.transaction_statement import *
from .receipt.resolution_statement import *
from .receipt.resolution_entry import *
from .receipt import *

# Config
from .config.catapult_config import *
from .config.catapult_upgrade import *
from .config import *

# Contract
from .contract.contract_info import *
from .contract import *

# Metadata
from .metadata.address_metadata_info import *
from .metadata.address_metadata import *
from .metadata.mosaic_metadata_info import *
from .metadata.mosaic_metadata import *
from .metadata.namespace_metadata_info import *
from .metadata.namespace_metadata import *
from .metadata.field import *
from .metadata.metadata_modification import *
from .metadata.metadata_type import *
from .metadata.metadata_modification_type import *
from .metadata.metadata_info import *
from .metadata import *

# Mosaic
from .mosaic.mosaic import *
from .mosaic.mosaic_id import *
from .mosaic.mosaic_info import *
from .mosaic.mosaic_name import *
from .mosaic.mosaic_nonce import *
from .mosaic.mosaic_properties import *
from .mosaic.mosaic_supply_type import *
from .mosaic.network_currency_mosaic import *
from .mosaic.network_harvest_mosaic import *
from .mosaic import *

# Namespace
from .namespace.address_alias import *
from .namespace.alias import *
from .namespace.alias_action_type import *
from .namespace.alias_type import *
from .namespace.empty_alias import *
from .namespace.mosaic_alias import *
from .namespace.namespace_id import *
from .namespace.namespace_info import *
from .namespace.namespace_name import *
from .namespace.namespace_type import *
from .namespace import *

# Node
from .node.node_info import *
from .node.node_time import *
from .node import *

# Transaction
from .transaction.account_link_transaction import *
from .transaction.account_property_modification import *
from .transaction.address_alias_transaction import *
from .transaction.aggregate_transaction_cosignature import *
from .transaction.aggregate_transaction_info import *
from .transaction.aggregate_transaction import *
from .transaction.alias_transaction import *
from .transaction.blockchain_upgrade_transaction import *
from .transaction.cosignature_signed_transaction import *
from .transaction.cosignature_transaction import *
from .transaction.deadline import *
from .transaction.hash_lock_transaction import *
from .transaction.hash_type import *
from .transaction.inner_transaction import *
from .transaction.link_action import *
from .transaction.lock_funds_transaction import *
from .transaction.message import *
from .transaction.message_type import *
from .transaction.modify_account_property_address_transaction import *
from .transaction.modify_account_property_entity_type_transaction import *
from .transaction.modify_account_property_mosaic_transaction import *
from .transaction.modify_account_property_transaction import *
from .transaction.modify_account_metadata_transaction import *
from .transaction.modify_mosaic_metadata_transaction import *
from .transaction.modify_namespace_metadata_transaction import *
from .transaction.modify_metadata_transaction import *
from .transaction.modify_multisig_account_transaction import *
from .transaction.mosaic_alias_transaction import *
from .transaction.mosaic_definition_transaction import *
from .transaction.mosaic_supply_change_transaction import *
from .transaction.multisig_cosignatory_modification import *
from .transaction.multisig_cosignatory_modification_type import *
from .transaction.network_config_transaction import *
from .transaction.plain_message import *
from .transaction.register_namespace_transaction import *
from .transaction.secret_lock_transaction import *
from .transaction.secret_proof_transaction import *
from .transaction.signed_transaction import *
from .transaction.sync_announce import *
from .transaction.transaction import *
from .transaction.transaction_announce_response import *
from .transaction.transaction_info import *
from .transaction.transaction_status import *
from .transaction.transaction_status_error import *
from .transaction.transaction_status_group import *
from .transaction.transaction_type import *
from .transaction.transaction_version import *
from .transaction.transfer_transaction import *
from .transaction import *

__all__ = (
    # Account
    account.__all__
    + account_info.__all__
    + account_meta.__all__
    + account_property.__all__
    + account_properties.__all__
    + account_names.__all__
    + address.__all__
    + multisig_account_graph_info.__all__
    + multisig_account_info.__all__
    + property_modification_type.__all__
    + property_type.__all__
    + public_account.__all__

    # Blockchain
    + block_info.__all__
    + block_type.__all__
    + blockchain_score.__all__
    + blockchain_storage_info.__all__
    + merkle_path_item.__all__
    + merkle_proof_info.__all__
    + network_type.__all__

    # Receipt
    + balance_change_receipt.__all__
    + balance_transfer_receipt.__all__
    + artifact_expiry_receipt.__all__
    + inflation_receipt.__all__
    + receipt.__all__
    + receipt_type.__all__
    + receipt_version.__all__
    + source.__all__
    + statements.__all__
    + transaction_statement.__all__
    + resolution_statement.__all__
    + resolution_entry.__all__

    # Config
    + catapult_config.__all__
    + catapult_upgrade.__all__

    # Metadata
    + address_metadata_info.__all__
    + address_metadata.__all__
    + mosaic_metadata_info.__all__
    + mosaic_metadata.__all__
    + namespace_metadata_info.__all__
    + namespace_metadata.__all__
    + metadata_type.__all__
    + metadata_modification_type.__all__
    + metadata_modification.__all__
    + metadata_info.__all__
    + field.__all__

    # Mosaic
    + mosaic.__all__
    + mosaic_id.__all__
    + mosaic_info.__all__
    + mosaic_name.__all__
    + mosaic_nonce.__all__
    + mosaic_properties.__all__
    + mosaic_supply_type.__all__
    + network_currency_mosaic.__all__
    + network_harvest_mosaic.__all__

    # Namespace
    + address_alias.__all__
    + alias.__all__
    + alias_action_type.__all__
    + alias_type.__all__
    + empty_alias.__all__
    + mosaic_alias.__all__
    + namespace_id.__all__
    + namespace_info.__all__
    + namespace_name.__all__
    + namespace_type.__all__

    # Node
    + node_info.__all__
    + node_time.__all__

    # Transaction
    + account_link_transaction.__all__
    + account_property_modification.__all__
    + address_alias_transaction.__all__
    + aggregate_transaction_cosignature.__all__
    + aggregate_transaction_info.__all__
    + aggregate_transaction.__all__
    + alias_transaction.__all__
    + blockchain_upgrade_transaction.__all__
    + cosignature_signed_transaction.__all__
    + cosignature_transaction.__all__
    + deadline.__all__
    + hash_lock_transaction.__all__
    + hash_type.__all__
    + inner_transaction.__all__
    + link_action.__all__
    + lock_funds_transaction.__all__
    + message.__all__
    + message_type.__all__
    + modify_account_property_address_transaction.__all__
    + modify_account_property_entity_type_transaction.__all__
    + modify_account_property_mosaic_transaction.__all__
    + modify_account_property_transaction.__all__
    + modify_account_metadata_transaction.__all__
    + modify_mosaic_metadata_transaction.__all__
    + modify_namespace_metadata_transaction.__all__
    + modify_metadata_transaction.__all__
    + modify_multisig_account_transaction.__all__
    + mosaic_alias_transaction.__all__
    + mosaic_definition_transaction.__all__
    + mosaic_supply_change_transaction.__all__
    + multisig_cosignatory_modification.__all__
    + multisig_cosignatory_modification_type.__all__
    + network_config_transaction.__all__
    + plain_message.__all__
    + register_namespace_transaction.__all__
    + secret_lock_transaction.__all__
    + secret_proof_transaction.__all__
    + signed_transaction.__all__
    + sync_announce.__all__
    + transaction.__all__
    + transaction_announce_response.__all__
    + transaction_info.__all__
    + transaction_status.__all__
    + transaction_status_error.__all__
    + transaction_status_group.__all__
    + transaction_type.__all__
    + transaction_version.__all__
    + transfer_transaction.__all__
)



"""
    sha3
    ====

    Implementation of the SHA3 cryptographic hash functions.

    License
    -------

    Copyright 2019 NEM

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""

from __future__ import annotations
import hashlib
import typing

# API


def sha3_224(data: typing.Optional[bytes] = None):
    """Returns a sha3_224 hash object; optionally initialized with a string."""

    if data is None:
        return hashlib.sha3_224()
    return hashlib.sha3_224(data)


def sha3_256(data: typing.Optional[bytes] = None):
    """Returns a sha3_256 hash object; optionally initialized with a string."""

    if data is None:
        return hashlib.sha3_256()
    return hashlib.sha3_256(data)


def sha3_384(data: typing.Optional[bytes] = None):
    """Returns a sha3_384 hash object; optionally initialized with a string."""

    if data is None:
        return hashlib.sha3_384()
    return hashlib.sha3_384(data)


def sha3_512(data: typing.Optional[bytes] = None):
    """Returns a sha3_512 hash object; optionally initialized with a string."""

    if data is None:
        return hashlib.sha3_512()
    return hashlib.sha3_512(data)



from xpxchain import models
from tests import harness


class TestBlockchainScore(harness.TestCase):

    @harness.randomize
    def test_valid(self, score: harness.U128):
        network_type = models.NetworkType.MIJIN_TEST
        model = models.BlockchainScore(score)
        dto = model.to_dto(network_type)
        self.assertEqual(model, models.BlockchainScore.create_from_dto(dto, network_type))

    @harness.randomize(score={'min_value': -1 << 32, 'max_value': -1})
    def test_invalid(self, score: int):
        network_type = models.NetworkType.MIJIN_TEST
        model = models.BlockchainScore(score)
        with self.assertRaises(ArithmeticError):
            model.to_dto(network_type)



"""
    mosaic_nonce
    ============

    Nonce for a mosaic.

    Nonces are arbitrary numbers used for cryptographic communications,
    used to avoid replay attacks.

    License
    -------

    Copyright 2019 NEM

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""

from __future__ import annotations
import os
import typing

from ..blockchain.network_type import OptionalNetworkType
from ... import util

__all__ = ['MosaicNonce']

RawNonceType = typing.Union[int, bytes, str]


def nonce_as_bytes(nonce: RawNonceType):
    """Convert nonce to underlying byte array."""
    if isinstance(nonce, int):
        return util.u32_to_catbuffer(nonce)
    elif isinstance(nonce, str):
        return util.unhexlify(nonce)
    elif isinstance(nonce, bytes):
        return nonce
    else:
        raise TypeError(f"Invalid nonce type, got {type(nonce)}.")


# TODO(ahuszagh) Change to an object, not an actual Model.
@util.inherit_doc
@util.dataclass(frozen=True)
class MosaicNonce(util.Model):
    """
    Nonce for a mosaic.

    :param nonce: Mosaic nonce.
    """

    nonce: bytes
    CATBUFFER_SIZE = util.U32_BYTES

    def __init__(self, nonce: typing.Union[int, str, bytes]) -> None:
        self._set('nonce', nonce_as_bytes(nonce))
        if len(self.nonce) != 4:
            raise ValueError("Nonce length is incorrect.")

    def __int__(self) -> int:
        return util.u32_from_catbuffer(self.nonce)

    @classmethod
    def create_random(cls, entropy=os.urandom):
        """
        Create new mosaic nonce from random bytes.

        :param entropy: (Optional) Callback to generate random bytes.
        """
        nonce: bytes = entropy(4)
        return cls(nonce)

    @classmethod
    def create_from_hex(cls, data: str):
        """
        Create mosaic nonce from hex-encoded nonce.

        :param data: Hex-encoded nonce data.
        """
        return cls(util.unhexlify(data))

    @classmethod
    def create_from_int(cls, nonce: int):
        """
        Create mosaic nonce from 32-bit integer.

        :param nonce: Nonce as 32-bit unsigned integer.
        """
        return cls(nonce)

    @classmethod
    def validate_dto(cls, data: int) -> bool:
        """Validate the data-transfer object."""
        return isinstance(data, int) and 0 <= data < (1 << 32)

    def to_dto(
        self,
        network_type: OptionalNetworkType = None,
    ) -> int:
        return int(self)

    @classmethod
    def create_from_dto(
        cls,
        data: int,
        network_type: OptionalNetworkType = None,
    ):
        # Rest api returns negative number but it should be unsigned. Anyway, the size
        # stays 4B so this mask should be OK
        data &= 0xFFFFFFFF

        if not cls.validate_dto(data):
            raise ValueError('Invalid data-transfer object.')

        return cls(data)

    def to_catbuffer(
        self,
        network_type: OptionalNetworkType = None,
        fee_strategy: typing.Optional[util.FeeCalculationStrategy] = util.FeeCalculationStrategy.MEDIUM,
    ) -> bytes:
        return util.u32_to_catbuffer(int(self))

    @classmethod
    def create_from_catbuffer(
        cls,
        data: bytes,
        network_type: OptionalNetworkType = None,
    ):
        size = cls.CATBUFFER_SIZE
        return cls(util.u32_from_catbuffer(data[:size]))



"""
    hash_type
    =========

    Validators for hashes.

    License
    -------

    Copyright 2019 NEM

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""

import enum
import string

from ... import util

__all__ = ['HashType']


class HashType(util.U8Mixin, util.EnumMixin, enum.IntEnum):
    """Enumerations for support hash types."""

    SHA3_256 = 0
    KECCAK_256 = 1
    HASH_160 = 2
    HASH_256 = 3

    def description(self) -> str:
        return DESCRIPTION[self]

    def hash_length(self) -> int:
        """Calculate the expected hash length."""
        return LENGTH[self]

    def validate(self, input: str) -> bool:
        """
        Validate hex-encoded hash as type.

        :param input: Hex-encoded hash data.
        """
        if all(i in string.hexdigits for i in input):
            return self.hash_length() == len(input)
        return False


DESCRIPTION = {
    HashType.SHA3_256: "SHA3-256 (default).",
    HashType.KECCAK_256: "Keccak-256 (ETH compatibility).",
    HashType.HASH_160: "SHA3-256 to RIPEMD-160 (BTC compatibility).",
    HashType.HASH_256: "SHA3-256 to SHA3-256 (BTC compatibility).",
}

LENGTH = {
    HashType.SHA3_256: 64,
    HashType.KECCAK_256: 64,
    HashType.HASH_160: 40,
    HashType.HASH_256: 64,
}



"""
    from_coroutine
    ==============

    RxPy extension to generate an observable from a coroutine.

    License
    -------

    Copyright 2019 NEM

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""

from __future__ import annotations
import asyncio
import rx
from rx.internal import extensionclassmethod


def from_coroutine(cls, coroutine, scheduler=None, loop=None):
    """
    Generate observable from an asynchronous coroutine.

    :param coroutine: Coroutine to wrap as observable.
    :param scheduler: (Optional) scheduler to determine the event loop.
    :param loop: (Optional) Event loop. Has higher precedence than scheduler.
    """

    # Get the current event loop.
    if loop is None:
        loop = asyncio.get_event_loop()
    elif scheduler is not None:
        loop = scheduler.loop

    task = loop.create_task(coroutine)
    return cls.from_future(task)


extensionclassmethod(rx.AnonymousObservable)(from_coroutine)
extensionclassmethod(rx.Observable)(from_coroutine)



"""
    abc
    ===

    Abstract base classes for NEM models.

    License
    -------

    Copyright 2019 NEM

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""

from __future__ import annotations
import enum
import typing

__all__ = [
    'AbstractMethodError',
    'Catbuffer',
    'DTO',
    'Object',
    'Model',
]

# TYPES

NetworkType = typing.TypeVar('NetworkType', bound=enum.IntEnum)
OptionalNetworkType = typing.Optional[NetworkType]
DTOType = typing.TypeVar('DTOType', bound='DTO')
CatbufferType = typing.TypeVar('CatbufferType', bound='Catbuffer')
ModelType = typing.TypeVar('ModelType', bound='Model')
DTOFormat = typing.TypeVar('DTOFormat', int, str, dict, list)

# EXCEPTIONS


class AbstractMethodError(NotImplementedError):
    pass


# MODELS

# typing.Generic defines a custom __new__, which causes enum
# to use it, rather than the data_type, as the constructor
# for the type. We can override the __new__ to call the correct
# type, however, we still return instances of that class with
# attributes of the data type built in. For example, for integers,
# we can still call __int__ to get the underlying value, however,
# __repr__, __eq__, and most other methods are overridden. This
# is especially problematic for str, since it effectively requires
# hacks to access the underlying value (which is true for everything).
# There are 2 possible solutions:
#   1. Define a new Generic base where `__new__`, `__init_subclass__`
#       are deleted from typing.Generic, so that mypy still checks types.
#   2. Define the base as Generic only when `typing.TYPE_CHECKING`.
#
# The latter is way less fragile, and if other tools start to break,
# the easy solution is to just remove `typing.Generic`, the optional
# type checks, rather than attempt to deconvolute complicated
# inheritance patterns. The condition should be on the invocation
# of mypy, since sphinx-doc will attempt to parse the actual source code.
#
# We define a new class, does not understand the base class from
# assignment as of '0.670'.
T = typing.TypeVar('T')
MYPY = False

if MYPY:

    class Object(typing.Generic[T]):
        """Transparent Generic subclass."""
        __slots__ = ()
else:

    class Object:
        """Class-indexable object that just returns objects."""
        __slots__ = ()

        def __class_getitem__(cls, params):
            return cls


class DTO(Object[DTOFormat]):
    """Future format of the DTO class."""

    __slots__ = ()

    @classmethod
    def validate_dto_required(
        cls: typing.Type[DTOType],
        data: dict,
        required_keys: set
    ) -> bool:
        """Validate the required keys of the DTO."""
        return set(data) & required_keys == required_keys

    @classmethod
    def validate_dto_all(
        cls: typing.Type[DTOType],
        data: dict,
        all_keys: set
    ) -> bool:
        """Validate all keys are known of the DTO."""
        if __debug__:
            # Only do in debug builds, since it might pose backwards-
            # compatibility issues in production.
            return not(set(data) - all_keys)
        return True

    @classmethod
    def validate_dto(
        cls: typing.Type[DTOType],
        data: DTOFormat
    ) -> bool:
        """
        Validate the DTO interchange format matches the object type.

        :param data: Data-transfer object.
        :return: If DTO was validated for type.
        """
        raise AbstractMethodError

    def to_dto(
        self: DTOType,
        network_type: OptionalNetworkType = None,
    ) -> DTOFormat:
        """
        Export model to DTO interchange format.

        :param network_type: (Optional) network type.
        :return: Model as DTO interchange format.
        """
        raise AbstractMethodError

    @classmethod
    def create_from_dto(
        cls: typing.Type[DTOType],
        data: DTOFormat,
        network_type: OptionalNetworkType = None,
    ) -> DTOType:
        """
        Load model from DTO interchange format.

        :param data: Data-transfer object.
        :param network_type: (Optional) network type.
        :return: Native model from DTO interchange format.
        """
        raise AbstractMethodError

    @classmethod
    def iter_to_dto(
        cls: typing.Type[DTOType],
        iterable: typing.Iterable[DTOType],
        network_type: OptionalNetworkType = None,
    ) -> typing.Generator[DTOFormat, None, None]:
        """
        Iteratively export models to DTO interchange format.

        :param iterable: Iterable yielding models.
        :param network_type: Network type.
        :return: Next model as DTO interchange format.
        """
        for model in iterable:
            yield model.to_dto(network_type)

    @classmethod
    def iter_from_dto(
        cls: typing.Type[DTOType],
        iterable: typing.Iterable[DTOFormat],
        network_type: OptionalNetworkType = None,
    ) -> typing.Generator[DTOType, None, None]:
        """
        Iteratively load models from DTO interchange format.

        :param iterable: Iterable yielding DTO objects.
        :param network_type: Network type.
        :return: Next native model from DTO interchange format.
        """
        for dto in iterable:
            yield cls.create_from_dto(dto, network_type)

    @classmethod
    def sequence_to_dto(
        cls: typing.Type[DTOType],
        sequence: typing.Sequence[DTOType],
        network_type: OptionalNetworkType = None,
    ) -> typing.Sequence[DTOFormat]:
        """
        Export sequence of models to DTO interchange format.

        :param sequence: Sequence of models.
        :param network_type: Network type.
        :return: List of models in DTO interchange format.
        """
        return list(cls.iter_to_dto(sequence, network_type))

    @classmethod
    def sequence_from_dto(
        cls: typing.Type[DTOType],
        sequence: typing.Sequence[DTOFormat],
        network_type: OptionalNetworkType = None,
    ) -> typing.Sequence[DTOType]:
        """
        Load list of models from DTO interchange format.

        :param sequence: Sequence of DTO objects.
        :param network_type: Network type.
        :return: List of native model from DTO interchange format.
        """
        return list(cls.iter_from_dto(sequence, network_type))


def is_same_classmethod(
    cls1: typing.Type,
    cls2: typing.Type,
    name: str,
) -> bool:
    """Check if two class methods are not the same instance."""

    clsmeth1 = getattr(cls1, name)
    clsmeth2 = getattr(cls2, name)
    return clsmeth1.__func__ is clsmeth2.__func__


class Catbuffer(Object):
    """
    Future format of the catbuffer class.

    Either `create_from_catbuffer` or `create_from_catbuffer_pair` needs
    to be specialized. If `CATBUFFER_SIZE` is present, `create_from_catbuffer`
    should be specialized, otherwise, `create_from_catbuffer_pair` should
    be specialized. If `CATBUFFER_SIZE` is present and neither function
    is specialized, it will lead to an `AssertionError`, or on optimized
    builds, a `RecursionError`.
    """

    __slots__ = ()
    CATBUFFER_SIZE: typing.ClassVar[int]

    def to_catbuffer(
        self: CatbufferType,
        network_type: OptionalNetworkType = None,
    ) -> bytes:
        """
        Export model to catbuffer interchange format.

        :param network_type: (Optional) network type.
        :return: Model as catbuffer interchange format.
        """
        raise AbstractMethodError

    @classmethod
    def create_from_catbuffer(
        cls: typing.Type[CatbufferType],
        data: bytes,
        network_type: OptionalNetworkType = None,
    ) -> CatbufferType:
        """
        Load model from catbuffer interchange format.

        If `CATBUFFER_SIZE` is defined, you should override this
        method, and `create_from_catbuffer_pair` will be automatically
        defined. Otherwise, override `create_from_catbuffer_pair`.

        :param network_type: (Optional) network type.
        :return: Native model from catbuffer interchange format.
        """
        assert not is_same_classmethod(cls, Catbuffer, 'create_from_catbuffer_pair')
        return cls.create_from_catbuffer_pair(data, network_type)[0]

    @classmethod
    def create_from_catbuffer_pair(
        cls: typing.Type[CatbufferType],
        data: bytes,
        network_type: OptionalNetworkType = None,
    ) -> typing.Tuple[CatbufferType, bytes]:
        """
        Load model from catbuffer interchange format.

        If `CATBUFFER_SIZE` is not defined, you should override this
        method, and `create_from_catbuffer` will be automatically defined.
        Otherwise, override `create_from_catbuffer`.

        :param network_type: (Optional) network type.
        :return: Model and remaining bytes leftover.
        """
        assert not is_same_classmethod(cls, Catbuffer, 'create_from_catbuffer')
        size = cls.CATBUFFER_SIZE
        inst = cls.create_from_catbuffer(data[:size], network_type)
        remaining = data[size:]
        return inst, remaining

    @classmethod
    def iter_to_catbuffer(
        cls: typing.Type[CatbufferType],
        iterable: typing.Iterable[CatbufferType],
        network_type,
    ) -> typing.Generator[bytes, None, None]:
        """
        Iteratively export models to catbuffer interchange format.

        :param iterable: Iterable yielding models.
        :param network_type: Network type.
        :return: Next model as catbuffer interchange format.
        """
        for model in iterable:
            yield model.to_catbuffer(network_type)

    @classmethod
    def iter_from_catbuffer(
        cls: typing.Type[CatbufferType],
        data: bytes,
        count: int,
        network_type,
    ) -> typing.Generator[CatbufferType, None, None]:
        """
        Iteratively load models from catbuffer interchange format.

        :param data: Raw data as catbuffer interchange format.
        :param count: Number of models to load.
        :param network_type: Network type.
        :return: Next native model from catbuffer interchange format.
        """
        for value, data in cls.iter_from_catbuffer_pair(data, count, network_type):
            yield value

    @classmethod
    def iter_from_catbuffer_pair(
        cls: typing.Type[CatbufferType],
        data: bytes,
        count: int,
        network_type,
    ) -> typing.Generator[typing.Tuple[CatbufferType, bytes], None, None]:
        """
        Iteratively load models from catbuffer interchange format.

        :param data: Raw data as catbuffer interchange format.
        :param count: Number of models to load.
        :param network_type: Network type.
        :return: Next native model and input bytes leftover.
        """
        for _ in range(count):
            value, data = cls.create_from_catbuffer_pair(data, network_type)
            yield value, data

    @classmethod
    def sequence_to_catbuffer(
        cls: typing.Type[CatbufferType],
        sequence: typing.Sequence[CatbufferType],
        network_type,
    ) -> bytes:
        """
        Export sequence of models to catbuffer interchange format.

        :param sequence: Sequence of models.
        :param network_type: Network type.
        :return: Concatenated bytes from each model exported to catbuffer.
        """
        return b''.join(cls.iter_to_catbuffer(sequence, network_type))

    @classmethod
    def sequence_from_catbuffer(
        cls: typing.Type[CatbufferType],
        data: bytes,
        count: int,
        network_type,
    ) -> typing.Sequence[CatbufferType]:
        """
        Load list of models from catbuffer interchange format.

        :param data: Raw data as catbuffer interchange format.
        :param count: Number of models to load.
        :param network_type: Network type.
        :return: List of native models.
        """
        return cls.sequence_from_catbuffer_pair(data, count, network_type)[0]

    @classmethod
    def sequence_from_catbuffer_pair(
        cls: typing.Type[CatbufferType],
        data: bytes,
        count: int,
        network_type,
    ) -> typing.Tuple[typing.Sequence[CatbufferType], bytes]:
        """
        Load list of models from catbuffer interchange format.

        :param data: Raw data as catbuffer interchange format.
        :param count: Number of models to load.
        :param network_type: Network type.
        :return: List of native models and input bytes leftover.
        """
        result = []
        iterable = cls.iter_from_catbuffer_pair(data, count, network_type)
        for value, data in iterable:
            result.append(value)
        return result, data


class Model(DTO, Catbuffer):
    """Base class for NEM models."""

    __slots__ = ()
